import re
from typing import List, Optional, Pattern

import click
from robot.api import Token

from robotidy.utils import normalize_name


def parse_csv(value):
    if not value:
        return []
    return [val for val in value.split(",")]


def str_to_bool(value):
    return value.lower() == "true"


def validate_regex(value: str) -> Optional[Pattern]:
    try:
        return re.compile(value)
    except re.error:
        raise ValueError(f"'{value}' is not a valid regular expression.") from None


class SkipConfig:
    """Skip configuration (global and for each transformer)."""

    # Following names will be taken from transformer config and provided to Skip class instead
    HANDLES = frozenset(
        {
            "skip_documentation",
            "skip_return_values",
            "skip_keyword_call",
            "skip_keyword_call_pattern",
            "skip_settings",
            "skip_arguments",
            "skip_setup",
            "skip_teardown",
            "skip_timeout",
            "skip_template",
            "skip_return_statement",
            "skip_tags",
            "skip_comments",
            "skip_block_comments",
        }
    )

    def __init__(
        self,
        documentation: bool = False,
        return_values: bool = False,
        keyword_call: Optional[List] = None,
        keyword_call_pattern: Optional[List] = None,
        settings: bool = False,
        arguments: bool = False,
        setup: bool = False,
        teardown: bool = False,
        timeout: bool = False,
        template: bool = False,
        return_statement: bool = False,
        tags: bool = False,
        comments: bool = False,
        block_comments: bool = False,
    ):
        self.documentation = documentation
        self.return_values = return_values
        self.keyword_call: List = keyword_call if keyword_call else []
        self.keyword_call_pattern: List = keyword_call_pattern if keyword_call_pattern else []
        self.settings = settings
        self.arguments = arguments
        self.setup = setup
        self.teardown = teardown
        self.timeout = timeout
        self.template = template
        self.return_statement = return_statement
        self.tags = tags
        self.comments = comments
        self.block_comments = block_comments

    def update_with_str_config(self, **kwargs):
        for name, value in kwargs.items():
            # find the value we're overriding and get its type from it
            original_value = self.__dict__[name]
            if isinstance(original_value, bool):
                self.__dict__[name] = str_to_bool(value)
            elif isinstance(original_value, list):
                parsed_list = parse_csv(value)
                self.__dict__[name].extend(parsed_list)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class Skip:
    """Defines global skip conditions for each transformer."""

    def __init__(self, skip_config: SkipConfig):
        self.return_values = skip_config.return_values
        self.documentation = skip_config.documentation
        self.comments = skip_config.comments
        self.block_comments = skip_config.block_comments
        self.keyword_call_names = {normalize_name(name) for name in skip_config.keyword_call}
        self.keyword_call_pattern = {validate_regex(pattern) for pattern in skip_config.keyword_call_pattern}
        self.any_keword_call = self.check_any_keyword_call()
        self.skip_settings = self.parse_skip_settings(skip_config)

    @staticmethod
    def parse_skip_settings(skip_config):
        settings = {"settings", "arguments", "setup", "teardown", "timeout", "template", "return_statement", "tags"}
        skip_settings = set()
        for setting in settings:
            if getattr(skip_config, setting):
                skip_settings.add(setting)
        return skip_settings

    def check_any_keyword_call(self):
        return self.keyword_call_names or self.keyword_call_pattern

    def keyword_call(self, node):
        if not getattr(node, "keyword", None) or not self.any_keword_call:
            return False
        normalized = normalize_name(node.keyword)
        if normalized in self.keyword_call_names:
            return True
        for pattern in self.keyword_call_pattern:
            if pattern.search(node.keyword):
                return True
        return False

    def setting(self, name):
        if not self.skip_settings:
            return False
        if "settings" in self.skip_settings:
            return True
        return name.lower() in self.skip_settings

    def comment(self, comment):
        if self.comments:
            return True
        if not self.block_comments:
            return False
        return comment.tokens and comment.tokens[0].type == Token.COMMENT


documentation_option = click.option(
    "--skip-documentation",
    is_flag=True,
    help="Skip formatting of documentation",
)
return_values_option = click.option(
    "--skip-return-values",
    is_flag=True,
    help="Skip formatting of return values",
)
keyword_call_option = click.option(
    "--skip-keyword-call",
    type=str,
    multiple=True,
    help="Keyword call name that should not be formatted",
)
keyword_call_pattern_option = click.option(
    "--skip-keyword-call-pattern",
    type=str,
    multiple=True,
    help="Keyword call name pattern that should not be formatted",
)
settings_option = click.option("--skip-settings", is_flag=True, help="Skip formatting of settings")
arguments_option = click.option(
    "--skip-arguments",
    is_flag=True,
    help="Skip formatting of arguments",
)
setup_option = click.option(
    "--skip-setup",
    is_flag=True,
    help="Skip formatting of setup",
)
teardown_option = click.option(
    "--skip-teardown",
    is_flag=True,
    help="Skip formatting of teardown",
)
timeout_option = click.option(
    "--skip-timeout",
    is_flag=True,
    help="Skip formatting of timeout",
)
template_option = click.option(
    "--skip-template",
    is_flag=True,
    help="Skip formatting of template",
)
return_option = click.option(
    "--skip-return",
    is_flag=True,
    help="Skip formatting of return statement",
)
tags_option = click.option(
    "--skip-tags",
    is_flag=True,
    help="Skip formatting of tags",
)
comments_option = click.option("--skip-comments", is_flag=True, help="Skip formatting of comments")
block_comments_option = click.option("--skip-block-comments", is_flag=True, help="Skip formatting of block comments")
option_group = {
    "name": "Skip formatting",
    "options": [
        "--skip-documentation",
        "--skip-return-values",
        "--skip-keyword-call",
        "--skip-keyword-call-pattern",
        "--skip-settings",
        "--skip-arguments",
        "--skip-setup",
        "--skip-teardown",
        "--skip-timeout",
        "--skip-template",
        "--skip-return",
        "--skip-tags",
        "--skip-comments",
        "--skip-block-comments",
    ],
}



from robocorp_ls_core.basic import implements
from robocorp_ls_core.constants import NULL
from robocorp_ls_core.protocols import IDirCache, check_implements, T, Sentinel
from robocorp_ls_core.robotframework_log import get_logger

from collections import namedtuple
from pathlib import Path
from typing import Any, Generic, Callable, Optional, TypeVar
import functools
import os
import collections


log = get_logger(__name__)


def instance_cache(func):
    """
    Use as decorator:

    class MyClass(object):

        @instance_cache
        def cache_this(self):
            ...

    Clear the cache with:

    instance = MyClass()
    instance.cache_this()
    instance.cache_this.cache_clear(instance)

    # set_lock(instance, lock) can be used to make the cache thread-safe.
    # i.e.:
    instance.cache_this.set_lock(instance, lock)

    """
    cache_key = "__cache__%s" % (func.__name__,)
    cache_key_lock = "__cache__%s_lock" % (func.__name__,)

    @functools.wraps(func)
    def new_func(self, *args, **kwargs):
        try:
            cache = getattr(self, "__instance_cache__")
        except:
            cache = {}
            setattr(self, "__instance_cache__", cache)

        lock = cache.get(cache_key_lock, NULL)
        with lock:
            try:
                func_cache = cache[cache_key]
            except KeyError:
                func_cache = cache[cache_key] = {}

            args_cache_key = None
            if args:
                args_cache_key = (args_cache_key, tuple(args))
            if kwargs:
                # We don't do that because if the caller uses args and then
                # later kwargs, we'd have to match the parameter to the position,
                # so, simplify for now and don't accept kwargs.
                raise AssertionError("Cannot currently deal with kwargs.")

            ret = func_cache.get(args_cache_key, Sentinel)
            if ret is Sentinel:
                ret = func(self, *args, **kwargs)
                func_cache[args_cache_key] = ret
            return ret

    def set_lock(self, lock):
        # Note that this will make sure we'll pre-create the lock so that
        # the __instance_cache__ is always there (so, there should be
        # no race condition for it).
        try:
            cache = getattr(self, "__instance_cache__")
        except:
            cache = {}
            setattr(self, "__instance_cache__", cache)

        cache[cache_key_lock] = lock

    def cache_clear(self):
        try:
            cache = getattr(self, "__instance_cache__")
        except:
            pass
        else:
            lock = cache.get(cache_key_lock, NULL)
            with lock:
                cache.pop(cache_key, None)

    new_func.cache_clear = cache_clear
    new_func.set_lock = set_lock
    return new_func


class DirCache(object):
    """
    To be used as:

    dir_cache = DirCache(cache_directory)
    dir_cache.store("some_key", 1)
    value = dir_cache.load("some_key", int) # Ok

    try:
        value = dir_cache.load("some_key", dict)
    except KeyError:
        ... error: value is not a dict

    try:
        dir_cache.load("key does not exist", dict)
    except KeyError:
        ... error, key does not exist
    """

    def __init__(self, cache_directory: str) -> None:
        self._cache_dir = cache_directory
        os.makedirs(self._cache_dir, exist_ok=True)

    def _encode_key(self, key: Any) -> str:
        import hashlib

        return hashlib.sha224(repr(key).encode("utf-8")).hexdigest()

    def _get_file_for_key(self, key: Any) -> str:
        key_encoded = self._encode_key(key)
        return os.path.join(self._cache_dir, key_encoded)

    @implements(IDirCache.store)
    def store(self, key, value):
        import json

        with open(self._get_file_for_key(key), "w") as stream:
            stream.write(json.dumps({"key": key, "value": value}))

    @implements(IDirCache.load)
    def load(self, key, expected_class):
        import json

        filename = self._get_file_for_key(key)
        if not os.path.exists(filename):
            raise KeyError(f"Key: {key} not found in cache.")

        try:
            with open(filename, "r") as stream:
                contents = json.loads(stream.read())
            value = contents["value"]

        except Exception:
            msg = f"Unable to load key: {key} from cache."
            log.debug(msg)
            raise KeyError(msg)

        if not isinstance(value, expected_class):
            raise KeyError(
                f"Unable to load key: {key} from cache (expected it to be a {expected_class} was {type(value)}."
            )
        return value

    @implements(IDirCache.discard)
    def discard(self, key):
        filename = self._get_file_for_key(key)
        try:
            os.remove(filename)
        except Exception:
            pass

    def __typecheckself__(self) -> None:
        _: IDirCache = check_implements(self)


CachedFileMTimeInfo = namedtuple("CachedFileMTimeInfo", "st_mtime, st_size, path")


class CachedFileInfo(Generic[T]):
    def __init__(self, file_path: Path, compute_value: Callable[[Path], T]):
        self._file_path = file_path
        self._mtime_info: Optional[CachedFileMTimeInfo] = self._get_mtime_cache_info(
            file_path
        )
        # Note that we only get the value after getting the mtime (so, the
        # constructor receives a callable and not the value so that the cache
        # has no risk of having stale values).
        self._value = compute_value(file_path)

    @property
    def value(self) -> T:
        return self._value

    @property
    def file_path(self) -> Path:
        return self._file_path

    def _get_mtime_cache_info(self, file_path: Path) -> Optional[CachedFileMTimeInfo]:
        """
        Cache based on the time/size of a given path.
        """
        try:
            stat = file_path.stat()
            return CachedFileMTimeInfo(stat.st_mtime, stat.st_size, str(file_path))
        except:
            # Probably removed in the meanwhile.
            log.exception(f"Unable to get mtime for: {file_path}")
            return None

    def is_cache_valid(self) -> bool:
        return self._mtime_info == self._get_mtime_cache_info(self.file_path)


KT = TypeVar("KT")  # Key type.
VT = TypeVar("VT")  # Value type.


class LRUCache(Generic[KT, VT]):
    """
    A cache which has a max size and a size to resize to when we'd go over the
    limit available in the LRU.

    It's possible to customize the maximum size of the cache and to give a
    custom weight for each entry.

    Note that this is not an exact-value capacity. It's possible that the
    capacity is exceeded in a case where an item added and the resize_to +
    item_size exceeds the max_size.

    This cache is NOT thread safe (users must lock if needed).

    It's designed so that it's a dict where the key points to a list with
        [key, value, access_time, entry_size]
    """

    def __init__(
        self,
        max_size: int = 100,
        resize_to: Optional[int] = None,
        get_size: Callable[[Any], int] = lambda obj: 1,
    ):
        assert max_size >= 1
        self.max_size = max_size

        if resize_to is None:
            resize_to = int(max_size * 0.7)
        assert resize_to < max_size

        self.resize_to = resize_to
        self._get_size = get_size

        self._dict: Any = collections.OrderedDict()
        self._current_size_usage = 0

    def clear(self) -> None:
        self._current_size_usage = 0
        self._dict.clear()

    @property
    def current_size_usage(self) -> int:
        return self._current_size_usage

    def __getitem__(self, key: KT) -> VT:
        item = self._dict[key]
        self._dict.move_to_end(key)
        return item[0]

    def __contains__(self, key: KT) -> bool:
        try:
            self[key]
            return True
        except KeyError:
            return False

    def __len__(self) -> int:
        return len(self._dict)

    def __delitem__(self, key: KT) -> None:
        entry = self._dict.pop(key)
        self._current_size_usage -= entry[-1]

    def pop(self, key: KT, default=Sentinel.SENTINEL) -> Optional[Any]:
        try:
            entry = self._dict.pop(key)
            self._current_size_usage -= entry[-1]
            return entry[0]
        except KeyError:
            if default is not Sentinel.SENTINEL:
                return default
            raise

    def __setitem__(self, key: KT, value: VT) -> None:
        item = self._dict.get(key)
        if item is None:
            new_size = self._get_size(value)

            if self._current_size_usage + new_size > self.max_size:
                # Note that we may exceed the size cache because we'll resize
                # to the target value without accounting for the size of
                # the new entry (and that's ok for this use case).
                self._resize_to()

            self._current_size_usage += new_size
            item = [value, new_size]
            self._dict[key] = item
        else:
            item[0] = value
            self._dict.move_to_end(key)

    def get(self, key: KT, default: Optional[Any] = None) -> Optional[Any]:
        try:
            return self[key]
        except KeyError:
            return default

    def _resize_to(self) -> None:
        while self._current_size_usage > self.resize_to:
            _key, entry = self._dict.popitem(last=False)
            self._current_size_usage -= entry[-1]



import sys
from typing import (
    Iterator,
    Optional,
    List,
    Tuple,
    Any,
    Union,
    Hashable,
    Callable,
    Dict,
    Iterable,
    Sequence,
)

import ast as ast_module
from robocorp_ls_core.lsp import Error, RangeTypedDict, PositionTypedDict
from robocorp_ls_core.robotframework_log import get_logger
from robotframework_ls.impl.protocols import (
    TokenInfo,
    NodeInfo,
    KeywordUsageInfo,
    ILibraryImportNode,
    IRobotToken,
    INode,
    IRobotVariableMatch,
    VarTokenInfo,
    IKeywordArg,
    VariableKind,
    AdditionalVarInfo,
)
from robotframework_ls.impl.text_utilities import normalize_robot_name
from robocorp_ls_core.basic import isinstance_name
import functools
import weakref
import threading
import typing
import itertools
from robotframework_ls.impl.robot_localization import LocalizationInfo


log = get_logger(__name__)


class _NodesProviderVisitor(ast_module.NodeVisitor):
    def __init__(self, on_node=lambda node: None):
        ast_module.NodeVisitor.__init__(self)
        self._stack = []
        self.on_node = on_node

    def generic_visit(self, node):
        self._stack.append(node)
        self.on_node(self._stack, node)
        ast_module.NodeVisitor.generic_visit(self, node)
        self._stack.pop()


class _PrinterVisitor(ast_module.NodeVisitor):
    def __init__(self, stream):
        ast_module.NodeVisitor.__init__(self)
        self._level = 0
        self._stream = stream

    def _replace_spacing(self, txt):
        curr_len = len(txt)
        delta = 80 - curr_len
        return txt.replace("*SPACING*", " " * delta)

    def generic_visit(self, node):
        # Note: prints line and col offsets 0-based (even if the ast is 1-based for
        # lines and 0-based for columns).
        self._level += 1
        try:
            indent = "  " * self._level
            node_lineno = node.lineno
            if node_lineno != -1:
                # Make 0-based
                node_lineno -= 1
            node_end_lineno = node.end_lineno
            if node_end_lineno != -1:
                # Make 0-based
                node_end_lineno -= 1
            self._stream.write(
                self._replace_spacing(
                    "%s%s *SPACING* (%s, %s) -> (%s, %s)\n"
                    % (
                        indent,
                        node.__class__.__name__,
                        node_lineno,
                        node.col_offset,
                        node_end_lineno,
                        node.end_col_offset,
                    )
                )
            )
            tokens = getattr(node, "tokens", [])
            for token in tokens:

                token_lineno = token.lineno
                if token_lineno != -1:
                    # Make 0-based
                    token_lineno -= 1

                self._stream.write(
                    self._replace_spacing(
                        "%s- %s, '%s' *SPACING* (%s, %s->%s)\n"
                        % (
                            indent,
                            token.type,
                            token.value.replace("\n", "\\n").replace("\r", "\\r"),
                            token_lineno,
                            token.col_offset,
                            token.end_col_offset,
                        )
                    )
                )

            ast_module.NodeVisitor.generic_visit(self, node)
        finally:
            self._level -= 1


MAX_ERRORS = 100


class _AbstractIndexer:
    def iter_indexed(self, clsname):
        pass

    @property
    def ast(self):
        return self._weak_ast()


class _FullIndexer(_AbstractIndexer):
    def __init__(self, weak_ast: "weakref.ref[ast_module.AST]"):
        self._weak_ast = weak_ast
        self._lock = threading.Lock()
        self._name_to_node_info_lst: Dict[str, List[NodeInfo]] = {}
        self._indexed_full = False

    def _index(self):
        with self._lock:
            if self._indexed_full:
                return

            ast = self._weak_ast()
            if ast is None:
                raise RuntimeError("AST already garbage collected.")

            for stack, node in _iter_nodes(ast):
                lst = self._name_to_node_info_lst.get(node.__class__.__name__)
                if lst is None:
                    lst = self._name_to_node_info_lst[node.__class__.__name__] = []

                lst.append(NodeInfo(tuple(stack), node))
            self._indexed_full = True

    def iter_indexed(self, clsname: str) -> Iterator[NodeInfo]:
        if not self._indexed_full:
            self._index()

        yield from iter(self._name_to_node_info_lst.get(clsname, ()))


class _SectionIndexer(_AbstractIndexer):
    """
    This is a bit smarter in that it can index only the parts we're interested
    in (so, to get the LibraryImport it won't iterate over the keywords to
    do the indexing).
    """

    INNER_INSIDE_TOP_LEVEL = {
        "LibraryImport": "SettingSection",
        "ResourceImport": "SettingSection",
        "VariablesImport": "SettingSection",
        "SuiteSetup": "SettingSection",
        "SuiteTeardown": "SettingSection",
        "TestTemplate": "SettingSection",
        # Not settings:
        "Keyword": "KeywordSection",
        "TestCase": "TestCaseSection",
        "Variable": "VariableSection",
    }

    TOP_LEVEL = {
        "SettingSection",
        "VariableSection",
        "TestCaseSection",
        "KeywordSection",
        "CommentSection",
    }

    def __init__(self, weak_ast):
        self._weak_ast = weak_ast
        self._lock = threading.Lock()
        self._first_level_name_to_node_info_lst: Dict[str, List[NodeInfo]] = {}

        # We always start by indexing the first level in this case (to get the sections
        # such as 'CommentSection', 'SettingSection', etc), which should be fast.

        ast = self._weak_ast()
        if ast is None:
            raise RuntimeError("AST already garbage collected.")

        for stack, node in _iter_nodes(ast, recursive=False):
            lst = self._first_level_name_to_node_info_lst.get(node.__class__.__name__)
            if lst is None:
                lst = self._first_level_name_to_node_info_lst[
                    node.__class__.__name__
                ] = []

            lst.append(NodeInfo(tuple(stack), node))

    def iter_indexed(self, clsname: str) -> Iterator[NodeInfo]:
        top_level = self.INNER_INSIDE_TOP_LEVEL.get(clsname)
        if top_level is not None:
            lst = self._first_level_name_to_node_info_lst.get(top_level)
            if lst is not None:
                for node_info in lst:
                    indexer = _obtain_ast_indexer(node_info.node)
                    yield from indexer.iter_indexed(clsname)
        else:
            if clsname in self.TOP_LEVEL:
                yield from iter(
                    self._first_level_name_to_node_info_lst.get(clsname, ())
                )
            else:
                # i.e.: We don't know what we should be getting, so, just check
                # everything...
                for lst in self._first_level_name_to_node_info_lst.values():
                    for node_info in lst:
                        indexer = _obtain_ast_indexer(node_info.node)
                        yield from indexer.iter_indexed(clsname)


class _ASTIndexer(_AbstractIndexer):
    def __init__(self, ast: ast_module.AST):
        self._weak_ast = weakref.ref(ast)
        self._is_root = ast.__class__.__name__ == "File"

        self._indexer: _AbstractIndexer
        if self._is_root:
            # Cache by sections
            self._indexer = _SectionIndexer(self._weak_ast)
        else:
            # Always cache fully
            self._indexer = _FullIndexer(self._weak_ast)

        self._additional_caches: Dict[Hashable, Tuple[Any, ...]] = {}

    def iter_cached(
        self, cache_key: Hashable, compute: Callable, *args
    ) -> Iterator[Any]:
        try:
            cached = self._additional_caches[cache_key]
        except KeyError:
            cached = tuple(compute(self, *args))
            self._additional_caches[cache_key] = cached

        yield from iter(cached)

    def iter_indexed(self, clsname: str) -> Iterator[NodeInfo]:
        return self._indexer.iter_indexed(clsname)


def _get_errors_from_tokens(node):
    for token in node.tokens:
        if token.type in (token.ERROR, token.FATAL_ERROR):
            start = (token.lineno - 1, token.col_offset)
            end = (token.lineno - 1, token.end_col_offset)
            error = Error(token.error, start, end)
            yield error


def _obtain_ast_indexer(ast):
    try:
        indexer = ast.__ast_indexer__
    except:
        indexer = ast.__ast_indexer__ = _ASTIndexer(ast)
    return indexer


def _convert_ast_to_indexer(func):
    @functools.wraps(func)
    def new_func(ast, *args, **kwargs):
        if hasattr(ast, "iter_indexed"):
            indexer = ast
        else:
            try:
                indexer = ast.__ast_indexer__
            except:
                indexer = ast.__ast_indexer__ = _ASTIndexer(ast)

        return func(indexer, *args, **kwargs)

    return new_func


def collect_errors(node) -> List[Error]:
    errors = []

    use_errors_attribute = "errors" in node.__class__._attributes

    for _stack, node in _iter_nodes(node, recursive=True):
        if node.__class__.__name__ == "Error":
            errors.extend(_get_errors_from_tokens(node))

        elif use_errors_attribute:
            node_errors = getattr(node, "errors", ())
            if node_errors:
                for error in node_errors:
                    errors.append(create_error_from_node(node, error, tokens=[node]))

        if len(errors) >= MAX_ERRORS:
            break

    return errors


def create_error_from_node(node, msg, tokens=None) -> Error:
    if tokens is None:
        tokens = node.tokens

    if not tokens:
        log.info("No tokens found when visiting: %s.", node.__class__)
        start = (0, 0)
        end = (0, 0)
    else:
        # line is 1-based and col is 0-based (make both 0-based for the error).
        start = (tokens[0].lineno - 1, tokens[0].col_offset)
        end = (tokens[-1].lineno - 1, tokens[-1].end_col_offset)

    error = Error(msg, start, end)
    return error


def print_ast(node, stream=None):
    if stream is None:
        stream = sys.stderr
    errors_visitor = _PrinterVisitor(stream)
    errors_visitor.visit(node)


def iter_sections(node):
    yield from iter(node.sections)


def find_keyword_section(node):
    for section in iter_sections(node):
        if isinstance_name(section, "KeywordSection"):
            return section
    return None


def is_keyword_section(node) -> bool:
    return isinstance_name(node, "KeywordSection")


def find_section(node, line: int) -> Optional[INode]:
    """
    :param line:
        0-based
    """
    last_section = None
    for section in iter_sections(node):
        # section.lineno is 1-based.
        if (section.lineno - 1) <= line:
            last_section = section

        else:
            return last_section

    return last_section


if typing.TYPE_CHECKING:
    # The INode has Robot Framework specific methods, but at runtime
    # we can just check the actual ast class.
    from typing import runtime_checkable, Protocol

    @runtime_checkable
    class _AST_CLASS(INode, Protocol):
        pass


else:
    # We know that the AST we're dealing with is the INode.
    # We can't use runtime_checkable on Python 3.7 though.
    _AST_CLASS = ast_module.AST


def get_local_variable_stack_and_node(
    stack: Sequence[INode],
) -> Tuple[Tuple[INode, ...], INode]:
    """
    Provides the stack to search local variables in (i.e.: the keyword/test case).

    Note that this requires a valid stack.
    """
    assert stack, "This method requires a valid stack."

    stack_lst: List[INode] = []
    for local_stack_node in reversed(stack):
        stack_lst.append(local_stack_node)
        if local_stack_node.__class__.__name__ in ("Keyword", "TestCase"):
            stack = tuple(stack_lst)
            break
    else:
        stack = (local_stack_node,)
        local_stack_node = stack[0]
    return stack, local_stack_node


def matches_stack(
    def_stack: Optional[Sequence[INode]], stack: Optional[Sequence[INode]]
) -> bool:
    """
    Note: just checks the stack, the source must be already validated at this point.
    """
    if stack is not None:
        if def_stack is None:
            return False

        if stack:
            if not def_stack:
                return False

            if stack[-1].lineno == def_stack[-1].lineno:
                return True

            # Not directly the same (we could be inside some for/while, so, let's
            # see if we can get the keyword/testcase from the stack).
            _, node1 = get_local_variable_stack_and_node(stack)
            _, node2 = get_local_variable_stack_and_node(def_stack)
            return node1.lineno == node2.lineno

    return True


def _iter_nodes(
    node, internal_stack: Optional[List[INode]] = None, recursive=True
) -> Iterator[Tuple[List[INode], INode]]:
    """
    :note: the yielded stack is actually always the same (mutable) list, so,
    clients that want to return it somewhere else should create a copy.
    """
    stack: List[INode]
    if internal_stack is None:
        stack = []
        if node.__class__.__name__ != "File":
            stack.append(node)
    else:
        stack = internal_stack

    if recursive:
        for _field, value in ast_module.iter_fields(node):
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, _AST_CLASS):
                        yield stack, item
                        stack.append(item)
                        yield from _iter_nodes(item, stack, recursive=True)
                        stack.pop()

            elif isinstance(value, _AST_CLASS):
                yield stack, value
                stack.append(value)

                yield from _iter_nodes(value, stack, recursive=True)

                stack.pop()
    else:
        # Not recursive
        for _field, value in ast_module.iter_fields(node):
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, _AST_CLASS):
                        yield stack, item

            elif isinstance(value, _AST_CLASS):
                yield stack, value


def _iter_nodes_reverse(node) -> Iterator[INode]:
    """
    Iterate over nodes in reverse order.
    """
    for field in reversed(node._fields):
        try:
            value = getattr(node, field)
        except AttributeError:
            continue

        if isinstance(value, list):
            for item in reversed(value):
                if isinstance(item, _AST_CLASS):
                    # i.e.: First go into
                    yield from _iter_nodes_reverse(item)
                    # And the provide current
                    yield item

        elif isinstance(value, _AST_CLASS):
            yield from _iter_nodes_reverse(value)


def iter_all_nodes_recursive(node: INode) -> Iterator[Tuple[List[INode], INode]]:
    """
    This function will iterate over all the nodes. Use only if there's no
    other way to implement it as iterating over all the nodes is slow...
    """
    yield from _iter_nodes(node)


def _iter_nodes_filtered_not_recursive(
    ast, accept_class: Union[Tuple[str, ...], str]
) -> Iterator[Tuple[list, Any]]:
    if not isinstance(accept_class, (list, tuple, set)):
        accept_class = (accept_class,)
    for stack, node in _iter_nodes(ast, recursive=False):
        if node.__class__.__name__ in accept_class:
            yield stack, node


def find_token(section, line, col) -> Optional[TokenInfo]:
    """
    :param section:
        The result from find_section(line, col), to pre-filter the nodes we may match.
    """
    for stack, node in _iter_nodes(section):
        try:
            tokens = node.tokens
        except AttributeError:
            continue

        if not tokens:
            continue

        if (tokens[-1].lineno - 1) < line:
            # i.e.: if the last node in the token is still before the
            # line we're searching, keep on going
            continue

        last_token = None
        for token in tokens:
            lineno = token.lineno - 1
            if lineno != line:
                continue

            if token.type == token.SEPARATOR:
                # For separator tokens, it must be entirely within the section
                # i.e.: if it's in the boundary for a word, we want the word,
                # not the separator.
                if token.col_offset < col < token.end_col_offset:
                    return TokenInfo(tuple(stack), node, token)

            elif token.type == token.EOL:
                # A trailing whitespace after a keyword should be part of
                # the keyword, not EOL.
                if token.col_offset <= col <= token.end_col_offset:
                    diff = col - token.col_offset
                    if last_token is not None and not token.value.strip():
                        eol_contents = token.value[:diff]
                        if len(eol_contents) <= 1:
                            token = _append_eol_to_prev_token(last_token, eol_contents)

                    return TokenInfo(tuple(stack), node, token)

            else:
                if token.col_offset <= col <= token.end_col_offset:
                    return TokenInfo(tuple(stack), node, token)

            last_token = token

    return None


def _find_subvar(stack, node, initial_token, col) -> Optional[VarTokenInfo]:
    for token, var_info in _tokenize_subvars(initial_token):
        if token.type == token.ARGUMENT:
            continue

        if token.col_offset <= col <= token.end_col_offset:
            if initial_token.type == initial_token.ASSIGN:
                token = copy_token_replacing(token, type=initial_token.ASSIGN)
            return VarTokenInfo(stack, node, token, var_info)

    from robotframework_ls.impl import robot_constants

    for p in robot_constants.VARIABLE_PREFIXES:
        if initial_token.value.startswith(p + "{"):
            if (
                initial_token.type == initial_token.ASSIGN
                and initial_token.value.endswith("}=")
            ):
                var_token = copy_token_with_subpart(initial_token, 2, -2)
            elif initial_token.value.endswith("}"):
                var_token = copy_token_with_subpart(initial_token, 2, -1)
            else:
                var_token = copy_token_with_subpart(
                    initial_token, 2, len(initial_token.value)
                )

            var_info = AdditionalVarInfo(p)
            return VarTokenInfo(stack, node, var_token, var_info)
    return None


def find_variable(section, line, col) -> Optional[VarTokenInfo]:
    """
    Finds the current variable token. Note that it won't include '{' nor '}'.
    The token may also be an empty token if we have a variable without contents.
    """

    token_info = find_token(section, line, col)
    if token_info is not None:
        stack = token_info.stack
        node = token_info.node
        token = token_info.token

        try:
            if (
                token.type == token.ARGUMENT
                and node.__class__.__name__ in CLASSES_WTH_EXPRESSION_ARGUMENTS
            ):
                for part, var_info in iter_expression_variables(token):
                    if part.type == token.VARIABLE:
                        if part.col_offset <= col <= part.end_col_offset:
                            return VarTokenInfo(
                                stack,
                                node,
                                part,
                                var_info.copy(
                                    context=AdditionalVarInfo.CONTEXT_EXPRESSION
                                ),
                            )

                if "$" in token.value:
                    char_in_token = col - token.col_offset
                    if char_in_token >= 0:
                        value_up_to_cursor = token.value[:char_in_token]
                        if value_up_to_cursor.endswith("$"):
                            # Empty variable at this point
                            from robot.api import Token

                            empty_token = Token(
                                type=token.VARIABLE,
                                value="",
                                lineno=token.lineno,
                                col_offset=col,
                            )

                            return VarTokenInfo(
                                stack,
                                node,
                                empty_token,
                                AdditionalVarInfo(
                                    "$",
                                    AdditionalVarInfo.CONTEXT_EXPRESSION,
                                ),
                            )

        except:
            log.exception("Unable to tokenize: %s", token)

        if "{" in token.value:
            parts = _tokenize_variables_even_when_invalid(token, col)
            if not parts:
                return None

            for part in parts:
                if part.type in (part.VARIABLE, part.ASSIGN):
                    if part.col_offset <= col <= part.end_col_offset:
                        return _find_subvar(
                            token_info.stack, token_info.node, part, col
                        )
            else:
                return None
    return None


def tokenize_variables_from_name(name):
    return tokenize_variables(create_token(name))  # May throw error if it's not OK.


def tokenize_variables(token: IRobotToken) -> Iterator[IRobotToken]:
    # May throw error if it's not OK.
    return iter(tuple(token.tokenize_variables()))


def _tokenize_variables_even_when_invalid(
    token: IRobotToken, col: int
) -> Iterator[IRobotToken]:
    """
    If Token.tokenize_variables() fails, this can still provide the variable under
    the given column by applying some heuristics to find open variables.
    """
    try:
        return iter(tuple(tokenize_variables(token)))
    except:
        pass

    # If we got here, it means that we weren't able to tokenize the variables
    # properly (probably some variable wasn't closed properly), so, let's do
    # a custom implementation for this use case.

    from robot.api import Token
    from robotframework_ls.impl.robot_constants import VARIABLE_PREFIXES

    diff = col - token.col_offset
    up_to_cursor = token.value[:diff]
    open_at = up_to_cursor.rfind("{")

    if open_at >= 1:
        if up_to_cursor[open_at - 1] in VARIABLE_PREFIXES:
            varname = [up_to_cursor[open_at - 1 :]]
            from_cursor = token.value[diff:]

            for c in from_cursor:
                if c in VARIABLE_PREFIXES or c.isspace() or c == "{":
                    break
                if c == "}":
                    varname.append(c)
                    break
                varname.append(c)

            return iter(
                [
                    Token(
                        type=token.VARIABLE,
                        value="".join(varname),
                        lineno=token.lineno,
                        col_offset=token.col_offset + open_at - 1,
                        error=token.error,
                    )
                ]
            )
    return iter(())


LIBRARY_IMPORT_CLASSES = ("LibraryImport",)
RESOURCE_IMPORT_CLASSES = ("ResourceImport",)
SETTING_SECTION_CLASSES = ("SettingSection",)


@_convert_ast_to_indexer
def iter_nodes(ast, accept_class: Union[Tuple[str, ...], str]) -> Iterator[NodeInfo]:
    """
    Note: always recursive.
    """
    if not isinstance(accept_class, (list, tuple, set)):
        accept_class = (accept_class,)

    for classname in accept_class:
        yield from ast.iter_indexed(classname)


def iter_all_nodes(ast, recursive=True) -> Iterator[NodeInfo]:
    """
    Note: use this *very* sparingly as no caching will take place
    (as all nodes need to be iterated).

    Use one of the filtered APIs whenever possible as those are cached
    by the type.
    """
    for stack, node in _iter_nodes(ast, recursive=recursive):
        yield NodeInfo(tuple(stack), node)


def is_library_node_info(node_info: NodeInfo) -> bool:
    return node_info.node.__class__.__name__ in LIBRARY_IMPORT_CLASSES


def is_resource_node_info(node_info: NodeInfo) -> bool:
    return node_info.node.__class__.__name__ in RESOURCE_IMPORT_CLASSES


def is_setting_section_node_info(node_info: NodeInfo) -> bool:
    return node_info.node.__class__.__name__ in SETTING_SECTION_CLASSES


@_convert_ast_to_indexer
def iter_library_imports(ast) -> Iterator[NodeInfo[ILibraryImportNode]]:
    cache_key = "iter_library_imports"
    yield from ast.iter_cached(cache_key, _iter_library_imports_uncached)


def _iter_library_imports_uncached(ast):
    try:
        from robot.api.parsing import LibraryImport  # noqa
    except ImportError:
        from robot.parsing.model.statements import LibraryImport  # noqa

    yield from ast.iter_indexed("LibraryImport")
    for keyword_usage_info in iter_keyword_usage_tokens(
        ast, collect_args_as_keywords=True
    ):
        if normalize_robot_name(keyword_usage_info.name) == "importlibrary":
            # Create a LibraryImport node based on the keyword usage.
            use_tokens = []
            iter_in = iter(keyword_usage_info.node.tokens)
            for token in iter_in:
                if token.type == token.KEYWORD:
                    # Skip the 'Import Library' keyword name.
                    break
            else:
                continue

            # Get the first non-separator token
            for token in iter_in:
                if token.type == token.SEPARATOR:
                    continue
                use_tokens.append(copy_token_replacing(token, type=token.NAME))
                break

            for token in iter_in:
                if token.type == token.ARGUMENT and token.value == "WITH NAME":
                    use_tokens.append(copy_token_replacing(token, type=token.WITH_NAME))
                    for token in iter_in:
                        if token.type == token.ARGUMENT:
                            use_tokens.append(
                                copy_token_replacing(token, type=token.NAME)
                            )
                        else:
                            use_tokens.append(token)

                else:
                    use_tokens.append(token)

            if use_tokens:
                node = LibraryImport(use_tokens)
                yield NodeInfo(keyword_usage_info.stack, node)


@_convert_ast_to_indexer
def iter_resource_imports(ast) -> Iterator[NodeInfo]:
    yield from ast.iter_indexed("ResourceImport")


@_convert_ast_to_indexer
def iter_variable_imports(ast) -> Iterator[NodeInfo]:
    yield from ast.iter_indexed("VariablesImport")


@_convert_ast_to_indexer
def iter_keywords(ast) -> Iterator[NodeInfo]:
    yield from ast.iter_indexed("Keyword")


@_convert_ast_to_indexer
def iter_variables(ast) -> Iterator[NodeInfo]:
    yield from ast.iter_indexed("Variable")


@_convert_ast_to_indexer
def iter_tests(ast) -> Iterator[NodeInfo]:
    yield from ast.iter_indexed("TestCase")


@_convert_ast_to_indexer
def iter_test_case_sections(ast) -> Iterator[NodeInfo]:
    yield from ast.iter_indexed("TestCaseSection")


@_convert_ast_to_indexer
def iter_setting_sections(ast) -> Iterator[NodeInfo]:
    yield from ast.iter_indexed("SettingSection")


@_convert_ast_to_indexer
def iter_indexed(ast, clsname) -> Iterator[NodeInfo]:
    yield from ast.iter_indexed(clsname)


def iter_keyword_arguments_as_str(ast, tokenize_keyword_name=False) -> Iterator[str]:
    """
    Provides the arguments with the full representation (use only for getting
    docs).

    May return strings as:
    ${my.store}=${my.load}[first][second]
    """
    for token in _iter_keyword_arguments_tokens(ast, tokenize_keyword_name):
        yield token.value


@_convert_ast_to_indexer
def _iter_keyword_arguments_tokens(
    ast, tokenize_keyword_name=False
) -> Iterator[IRobotToken]:
    """
    This API provides tokens as they are.

    i.e.: it returns tokens as: ${my.store}=${my.load}[first][second]

    So, this is internal as the outer world is tailored to
    dealing with what's actually needed.

    Note that it may return Token.ARGUMENT types (if defined in [Argument]) or
    Token.VARIABLE (if defined in keyword name).
    """
    for node_info in ast.iter_indexed("Arguments"):
        for token in node_info.node.tokens:
            if token.type == token.ARGUMENT:
                yield token

    if tokenize_keyword_name:
        from robot.api import Token

        ast_node = ast.ast
        if ast_node.__class__.__name__ == "Keyword":
            keyword_name = ast_node.header.get_token(Token.KEYWORD_NAME)
            if keyword_name:
                try:
                    tokenized_vars = keyword_name.tokenize_variables()
                except:
                    pass
                else:
                    for tok in tokenized_vars:
                        if tok.type == Token.VARIABLE:
                            yield tok


def iter_keyword_arguments_as_tokens(ast) -> Iterator[IRobotToken]:
    """
    API tailored at getting variables from keyword arguments.

    It converts an argument such as:
    "[Arguments]    ${my.store}=${my.load}[first][second]"

    and yields something as:

    "my.store"

    It may also convert keyword arguments in the keyword name such as:

    "Today is ${date:\d{4}-\d{2}-\d{2}}"

    to "date"
    """
    from robotframework_ls.impl.variable_resolve import iter_robot_variable_matches
    from robot.api import Token

    for initial_token in _iter_keyword_arguments_tokens(
        ast, tokenize_keyword_name=True
    ):
        for robot_match, _relative_index in iter_robot_variable_matches(
            initial_token.value
        ):
            i = initial_token.value.find(robot_match.base, robot_match.start)

            t = Token(
                initial_token.type,
                robot_match.base,
                initial_token.lineno,
                initial_token.col_offset + i,
                initial_token.error,
            )
            if initial_token.type == initial_token.VARIABLE:
                i = t.value.find(":")
                if i > 0:
                    t = copy_token_with_subpart(t, 0, i)
            yield t
            break  # Go to next arg.


def iter_keyword_arguments_as_kwarg(
    ast, tokenize_keyword_name=False
) -> Iterator[IKeywordArg]:

    from robotframework_ls.impl.robot_specbuilder import KeywordArg

    for token in _iter_keyword_arguments_tokens(ast, tokenize_keyword_name):
        yield KeywordArg(token.value)


def is_deprecated(ast) -> bool:
    from robotframework_ls.impl.text_utilities import has_deprecated_text

    docs = get_documentation_raw(ast)
    return has_deprecated_text(docs)


def get_documentation_raw(ast: INode) -> str:
    iter_in: Iterator[INode]
    if ast.__class__.__name__ == "File":
        # Handle the case where the File is given (docs must be gotten from
        # the *** Settings *** in this case).
        iter_in = (node_info.node for node_info in iter_setting_sections(ast))
    else:
        iter_in = iter((ast,))

    doc: List[str] = []
    last_line: List[str] = []

    last_token = None
    for ast_node in iter_in:
        for _stack, node in _iter_nodes_filtered_not_recursive(
            ast_node, accept_class="Documentation"
        ):
            for token in node.tokens:
                if last_token is not None and last_token.lineno != token.lineno:
                    doc.extend(last_line)
                    del last_line[:]

                last_token = token

                if token.type in (token.CONTINUATION, token.DOCUMENTATION):
                    # Ignore anything before a continuation.
                    del last_line[:]
                    continue

                last_line.append(token.value)
            else:
                # Last iteration
                doc.extend(last_line)

        if doc:
            # Case with multiple Setting sections
            break

    ret = "".join(doc).strip()
    return ret


def get_documentation_as_markdown(ast) -> str:
    documentation = get_documentation_raw(ast)
    if not documentation:
        return documentation
    try:
        from robotframework_ls import robot_to_markdown

        return robot_to_markdown.convert(documentation)
    except:
        log.exception("Error converting to markdown: %s", documentation)
        return documentation


KEYWORD_SET_LOCAL_TO_VAR_KIND = {
    normalize_robot_name("Set Local Variable"): VariableKind.LOCAL_SET_VARIABLE,
}

KEYWORD_SET_GLOBAL_TO_VAR_KIND = {
    normalize_robot_name("Set Task Variable"): VariableKind.TASK_SET_VARIABLE,
    normalize_robot_name("Set Test Variable"): VariableKind.TEST_SET_VARIABLE,
    normalize_robot_name("Set Suite Variable"): VariableKind.SUITE_SET_VARIABLE,
    normalize_robot_name("Set Global Variable"): VariableKind.GLOBAL_SET_VARIABLE,
}


KEYWORD_SET_ENV_TO_VAR_KIND = {
    normalize_robot_name("Set Environment Variable"): VariableKind.ENV_SET_VARIABLE,
}


@_convert_ast_to_indexer
def iter_local_assigns(ast) -> Iterator[VarTokenInfo]:
    from robot.api import Token

    for clsname, assign_token_type in (
        ("KeywordCall", Token.ASSIGN),
        ("ForHeader", Token.VARIABLE),  # RF 4+
        ("ForLoopHeader", Token.VARIABLE),  # RF 3
        ("ExceptHeader", Token.VARIABLE),
        ("InlineIfHeader", Token.ASSIGN),
    ):
        for node_info in ast.iter_indexed(clsname):
            node = node_info.node
            for token in node.tokens:
                if token.type == assign_token_type:
                    value = token.value

                    i = value.find("{")
                    j = value.rfind("}")
                    if i != -1 and j != -1 and i >= 1:
                        new_value = value[i + 1 : j]
                        token = Token(
                            type=token.type,
                            value=new_value,
                            lineno=token.lineno,
                            col_offset=token.col_offset + i + 1,
                            error=token.error,
                        )

                        yield VarTokenInfo(node_info.stack, node, token, value[0])


_FIXTURE_CLASS_NAMES = (
    "Setup",
    "Teardown",
    "SuiteSetup",
    "SuiteTeardown",
    "TestSetup",
    "TestTeardown",
)

_CLASSES_WITH_ARGUMENTS_AS_KEYWORD_CALLS_AS_TUPLE = _FIXTURE_CLASS_NAMES + (
    "TestTemplate",
    "Template",
)

_CLASSES_KEYWORDS_AND_OTHERS_WITH_ARGUMENTS_AS_KEYWORD_CALLS = (
    "KeywordCall",
) + _CLASSES_WITH_ARGUMENTS_AS_KEYWORD_CALLS_AS_TUPLE

CLASSES_WITH_ARGUMENTS_AS_KEYWORD_CALLS_AS_SET = frozenset(
    _CLASSES_WITH_ARGUMENTS_AS_KEYWORD_CALLS_AS_TUPLE
)

CLASSES_WTH_EXPRESSION_ARGUMENTS = (
    "IfHeader",
    "ElseIfHeader",
    "WhileHeader",
    "InlineIfHeader",
)


def _tokenize_subvars(
    initial_token: IRobotToken,
) -> Iterator[Tuple[IRobotToken, AdditionalVarInfo]]:
    if "{" not in initial_token.value:
        return

    for tok, var_info in _tokenize_subvars_tokens(initial_token):
        if tok.type in (tok.ARGUMENT, tok.VARIABLE):
            yield tok, var_info


def _tokenize_subvars_tokens(
    initial_token: IRobotToken,
    op_type: str = "variableOperator",
    var_type: Optional[str] = None,
) -> Iterator[Tuple[IRobotToken, AdditionalVarInfo]]:

    from robot.api import Token

    if var_type is None:
        var_type = Token.ARGUMENT

    if "{" not in initial_token.value:
        return

    if initial_token.value.startswith("{") and initial_token.value.endswith("}"):
        # i.e.: We're dealing with an expression.
        first, second, third = split_token_in_3(
            initial_token,
            op_type,
            var_type,
            op_type,
            1,
            -1,
        )
        yield first, AdditionalVarInfo()
        yield from iter_expression_tokens(second)
        yield third, AdditionalVarInfo()
        return

    robot_match_generator = RobotMatchTokensGenerator(initial_token, var_type)
    from robotframework_ls.impl.variable_resolve import iter_robot_variable_matches

    for robot_match, relative_index in iter_robot_variable_matches(initial_token.value):

        yield from robot_match_generator.gen_default_type(
            relative_index + len(robot_match.before)
        )

        yield from robot_match_generator.gen_tokens_from_robot_match(
            robot_match, relative_index, var_type=var_type
        )

    yield from robot_match_generator.gen_default_type(len(initial_token.value))


def _is_store_keyword(node):
    from robot.api import Token

    keyword_name_tok = node.get_token(Token.KEYWORD)
    if not keyword_name_tok:
        return False
    normalized = normalize_robot_name(keyword_name_tok.value)
    return (
        normalized in KEYWORD_SET_LOCAL_TO_VAR_KIND
        or normalized in KEYWORD_SET_GLOBAL_TO_VAR_KIND
    )


def _add_match(found: set, tok: IRobotToken) -> bool:
    """
    Helper to avoid returning 2 matches in the same position if 2 different
    heuristics overlap what they can return.
    """
    key = tok.col_offset, tok.lineno
    if key in found:
        return False
    found.add(key)
    return True


@_convert_ast_to_indexer
def iter_variable_references(ast) -> Iterator[VarTokenInfo]:
    from robotframework_ls.impl.ast_utils_keyword_usage import (
        obtain_keyword_usage_handler,
    )

    # TODO: This right now makes everything globally, we should have 2 versions,
    # one to resolve references which are global and another to resolve references
    # just inside some scope when dealing with local variables.
    # Right now we're not very smart and even if a variable is local we'll reference
    # global variables...

    # Note: we collect only the references, not the definitions here!
    found: set = set()
    for clsname in (
        "KeywordCall",
        "LibraryImport",
        "ResourceImport",
        "TestTimeout",
        "Variable",
        "ForHeader",  # RF 4+
        "ForLoopHeader",  # RF 3
    ) + _FIXTURE_CLASS_NAMES:
        for node_info in ast.iter_indexed(clsname):
            stack = node_info.stack
            node = node_info.node
            token = None
            arg_i = 0
            for token in node.tokens:
                try:
                    if token.type == token.ARGUMENT:
                        arg_i += 1
                        if arg_i == 1 and clsname == "KeywordCall":
                            if _is_store_keyword(node_info.node):
                                continue

                    if token.type == token.KEYWORD:
                        # Keyword calls may also have variables (unfortunately
                        # RF doesn't tokenize it with that type, so, we have
                        # to apply a workaround to change the type).
                        token = copy_token_replacing(token, type=token.NAME)

                    if token.type in (token.ARGUMENT, token.NAME):
                        for tok in tokenize_variables(token):
                            if tok.type == token.VARIABLE:
                                # We need to check for inner variables (as in
                                # this case we validate those).
                                for t, var_info in _tokenize_subvars(tok):
                                    if t.type != token.VARIABLE:
                                        continue
                                    if not _add_match(found, t):
                                        continue

                                    yield VarTokenInfo(stack, node, t, var_info)

                except:
                    log.exception("Unable to tokenize: %s", token)

    for node_info in _iter_node_info_which_may_have_usage_info(ast):
        stack = node_info.stack
        node = node_info.node
        keyword_usage_handler = obtain_keyword_usage_handler(stack, node)
        if keyword_usage_handler is not None:
            for usage_info in keyword_usage_handler.iter_keyword_usages_from_node():

                arg_i = 0
                for token in usage_info.node.tokens:
                    if token.type == token.ARGUMENT:
                        arg_i += 1
                        if arg_i == 1:
                            if _is_store_keyword(usage_info.node):
                                continue

                        next_tok_type = keyword_usage_handler.get_token_type(token)
                        if next_tok_type == keyword_usage_handler.EXPRESSION:
                            for tok, var_info in iter_expression_variables(token):
                                if tok.type == token.VARIABLE:
                                    if not _add_match(found, tok):
                                        continue
                                    yield VarTokenInfo(stack, node, tok, var_info)

    for clsname in CLASSES_WTH_EXPRESSION_ARGUMENTS:
        for node_info in ast.iter_indexed(clsname):
            stack = node_info.stack
            node = node_info.node
            token = None

            for token in node.tokens:
                try:
                    if token.type == token.ARGUMENT:
                        for tok, var_info in iter_expression_variables(token):
                            if tok.type == token.VARIABLE:
                                if not _add_match(found, tok):
                                    continue
                                yield VarTokenInfo(stack, node, tok, var_info)
                except:
                    log.exception("Unable to tokenize: %s", token)

    iter_keyword_node_info: Iterator[NodeInfo]
    if isinstance_name(ast.ast, "Keyword"):
        iter_keyword_node_info = iter((NodeInfo((ast.ast,), ast.ast),))
    else:
        iter_keyword_node_info = ast.iter_indexed("Keyword")

    for node_info in iter_keyword_node_info:
        stack = [node_info.node]
        for token in _iter_keyword_arguments_tokens(
            node_info.node, tokenize_keyword_name=True
        ):
            iter_in = _tokenize_subvars(token)

            try:
                # The first one is the variable store (the other is the
                # variable load on a default argument)
                # We are only interested in the second in this API.
                next(iter_in)
            except StopIteration:
                continue

            for t, varid in iter_in:
                if t.type != t.VARIABLE:
                    continue
                if not _add_match(found, t):
                    continue
                yield VarTokenInfo(stack, node, t, varid)


@_convert_ast_to_indexer
def iter_keyword_usage_tokens(
    ast, collect_args_as_keywords: bool
) -> Iterator[KeywordUsageInfo]:
    """
    Iterates through all the places where a keyword name is being used, providing
    the stack, node, token and name.
    """

    cache_key = ("iter_keyword_usage_tokens", collect_args_as_keywords)
    yield from ast.iter_cached(
        cache_key, _iter_keyword_usage_tokens_uncached, collect_args_as_keywords
    )


def _same_line_col(tok1: IRobotToken, tok2: IRobotToken):
    return tok1.lineno == tok2.lineno and tok1.col_offset == tok2.col_offset


def is_keyword_usage_node(ast):
    return (
        ast.__class__.__name__
        in _CLASSES_KEYWORDS_AND_OTHERS_WITH_ARGUMENTS_AS_KEYWORD_CALLS
    )


@_convert_ast_to_indexer
def _iter_node_info_which_may_have_usage_info(ast):
    for clsname in _CLASSES_KEYWORDS_AND_OTHERS_WITH_ARGUMENTS_AS_KEYWORD_CALLS:
        yield from ast.iter_indexed(clsname)


def _iter_keyword_usage_tokens_uncached(
    ast, collect_args_as_keywords: bool
) -> Iterator[KeywordUsageInfo]:
    from robotframework_ls.impl.ast_utils_keyword_usage import (
        obtain_keyword_usage_handler,
    )

    for node_info in _iter_node_info_which_may_have_usage_info(ast):
        stack = node_info.stack
        node = node_info.node
        keyword_usage_handler = obtain_keyword_usage_handler(
            stack, node, recursive=collect_args_as_keywords
        )
        if keyword_usage_handler is not None:
            yield from keyword_usage_handler.iter_keyword_usages_from_node()


def create_keyword_usage_info_from_token(
    stack: Tuple[INode, ...], node: INode, token: IRobotToken
) -> Optional[KeywordUsageInfo]:
    """
    If this is a keyword usage node, return information on it, otherwise,
    returns None.

    Note that it should return the keyword usage for the whole keyword call
    if we're in an argument that isn't itself a keyword call.
    """
    if token.type == token.ARGUMENT:
        from robotframework_ls.impl.ast_utils_keyword_usage import (
            obtain_keyword_usage_for_token,
        )

        return obtain_keyword_usage_for_token(stack, node, token)

    from robotframework_ls.impl.ast_utils_keyword_usage import (
        _create_root_keyword_usage_info,
    )

    return _create_root_keyword_usage_info(stack, node)


def get_keyword_name_token(
    stack: Tuple[INode, ...],
    node: INode,
    token: IRobotToken,
    accept_only_over_keyword_name: bool = True,
) -> Optional[IRobotToken]:
    """
    If the given token is a keyword call name, return the token, otherwise return None.

    :param accept_only_over_keyword_name:
        If True we'll only accept the token if it's over the keyword name.
        If False we'll accept the token even if it's over a keyword parameter.

    :note: this goes hand-in-hand with iter_keyword_usage_tokens.
    """
    if token.type == token.KEYWORD or (
        token.type == token.NAME
        and node.__class__.__name__ in CLASSES_WITH_ARGUMENTS_AS_KEYWORD_CALLS_AS_SET
    ):
        locinfo = get_localization_info_from_model(stack[0])
        return _strip_token_bdd_prefix(token, locinfo)[1]

    if token.type == token.ARGUMENT and not token.value.strip().endswith("}"):
        from robotframework_ls.impl.ast_utils_keyword_usage import (
            obtain_keyword_usage_for_token,
        )

        keyword_usage = obtain_keyword_usage_for_token(stack, node, token)
        if keyword_usage is not None:
            if accept_only_over_keyword_name:
                if _same_line_col(keyword_usage.token, token):
                    return token
            else:
                return keyword_usage.token
    return None


def get_library_import_name_token(node, token: IRobotToken) -> Optional[IRobotToken]:
    """
    If the given ast node is a library import and the token is its name, return
    the name token, otherwise, return None.
    """

    if (
        token.type == token.NAME
        and isinstance_name(node, "LibraryImport")
        and node.name == token.value  # I.e.: match the name, not the alias.
    ):
        return token
    return None


def get_resource_import_name_token(node, token: IRobotToken) -> Optional[IRobotToken]:
    """
    If the given ast node is a library import and the token is its name, return
    the name token, otherwise, return None.
    """

    if (
        token.type == token.NAME
        and isinstance_name(node, "ResourceImport")
        and node.name == token.value  # I.e.: match the name, not the alias.
    ):
        return token
    return None


def get_variables_import_name_token(ast, token):
    """
    If the given ast node is a variables import and the token is its name, return
    the name token, otherwise, return None.
    """

    if (
        token.type == token.NAME
        and isinstance_name(ast, "VariablesImport")
        and ast.name == token.value  # I.e.: match the name, not the alias.
    ):
        return token
    return None


def _copy_of_node_replacing_token(node, token, token_type):
    """
    Workaround to create a new version of the same node but with the first
    occurrence of a token of the given type changed to another token.
    """
    new_tokens = list(node.tokens)
    for i, t in enumerate(new_tokens):
        if t.type == token_type:
            new_tokens[i] = token
            break
    return node.__class__(new_tokens)


def _strip_node_and_token_bdd_prefix(
    stack: List[INode],
    node: INode,
    token_type: str,
) -> Tuple[str, INode, Optional[IRobotToken]]:
    """
    This is a workaround because the parsing does not separate a BDD prefix from
    the keyword name. If the parsing is improved to do that separation in the future
    we can stop doing this.
    """
    original_token = node.get_token(token_type)
    if original_token is None:
        return "", node, None

    locinfo = get_localization_info_from_model(stack[0])
    prefix, token = _strip_token_bdd_prefix(original_token, locinfo)
    if token is original_token:
        # i.e.: No change was done.
        return prefix, node, token
    return prefix, _copy_of_node_replacing_token(node, token, token_type), token


def _strip_token_bdd_prefix(
    token: IRobotToken, locinfo: LocalizationInfo
) -> Tuple[str, IRobotToken]:
    """
    This is a workaround because the parsing does not separate a BDD prefix from
    the keyword name. If the parsing is improved to do that separation in the future
    we can stop doing this.

    :return: the prefix and a new token with the bdd prefix stripped or the
    original token passed (if no prefix was detected).
    """
    from robot.api import Token

    assert token is not None

    text = token.value.lower()
    for prefix in locinfo.iter_bdd_prefixes_on_read():
        if text.startswith(prefix):
            try:
                next_is_space = text[len(prefix)] == " "
            except IndexError:
                continue

            if next_is_space:
                new_name = token.value[len(prefix) + 1 :]
                return prefix, Token(
                    type=token.type,
                    value=new_name,
                    lineno=token.lineno,
                    col_offset=token.col_offset + len(prefix) + 1,
                    error=token.error,
                )
    return "", token


def _append_eol_to_prev_token(last_token, eol_token_contents):
    from robot.api import Token

    new_value = last_token.value + eol_token_contents

    return Token(
        type=last_token.type,
        value=new_value,
        lineno=last_token.lineno,
        col_offset=last_token.col_offset,
        error=last_token.error,
    )


def copy_token_replacing(token, **kwargs):
    from robot.api import Token

    new_kwargs = {
        "type": token.type,
        "value": token.value,
        "lineno": token.lineno,
        "col_offset": token.col_offset,
        "error": token.error,
    }
    new_kwargs.update(kwargs)
    return Token(**new_kwargs)


def copy_token_with_subpart(token, start, end):
    from robot.api import Token

    return Token(
        type=token.type,
        value=token.value[start:end],
        lineno=token.lineno,
        col_offset=token.col_offset + start,
        error=token.error,
    )


def create_range_from_token(token: IRobotToken) -> RangeTypedDict:
    start: PositionTypedDict = {"line": token.lineno - 1, "character": token.col_offset}
    end: PositionTypedDict = {
        "line": token.lineno - 1,
        "character": token.end_col_offset,
    }
    taken_range: RangeTypedDict = {"start": start, "end": end}
    return taken_range


def create_range_from_tokens(
    token: IRobotToken, end_token: IRobotToken
) -> RangeTypedDict:
    start: PositionTypedDict = {"line": token.lineno - 1, "character": token.col_offset}
    end: PositionTypedDict = {
        "line": end_token.lineno - 1,
        "character": end_token.end_col_offset,
    }
    taken_range: RangeTypedDict = {"start": start, "end": end}
    return taken_range


def create_range_from_node(node: INode, accept_empty=True) -> Optional[RangeTypedDict]:
    first_token = None
    last_found_tokens = None

    # Find the first token
    for n in itertools.chain(
        iter((node,)), (x[1] for x in _iter_nodes(node, recursive=True))
    ):
        try:
            last_found_tokens = n.tokens
        except AttributeError:
            continue

        if not last_found_tokens:
            continue

        if accept_empty:
            first_token = next(iter(last_found_tokens))
        else:
            for t in last_found_tokens:
                if t.type in (t.EOL, t.EOS, t.SEPARATOR):
                    continue
                first_token = t
                break

        if first_token is not None:
            break

    if first_token is None:
        return None

    # Ok, we found the first one, now, let's find the last one.
    last_token = None
    last_found_tokens = None

    for n in itertools.chain(iter((node,)), _iter_nodes_reverse(node)):
        try:
            last_found_tokens = n.tokens
        except AttributeError:
            continue

        if not last_found_tokens:
            continue

        if accept_empty:
            last_token = next(reversed(last_found_tokens))
        else:
            for t in reversed(last_found_tokens):
                if t.type in (t.EOL, t.EOS, t.SEPARATOR):
                    continue
                last_token = t
                break

        if last_token is not None:
            break

    if last_token is None:
        return None

    return create_range_from_tokens(first_token, last_token)


def create_token(name):
    from robot.api import Token

    return Token(Token.NAME, name)


def convert_variable_match_base_to_token(
    token: IRobotToken, variable_match: IRobotVariableMatch
):
    from robot.api import Token

    base = variable_match.base
    assert base is not None
    s = variable_match.string
    if not base:
        base_i = s.find("{") + 1
    else:
        base_i = s.find(base)

    return Token(
        type=token.type,
        value=variable_match.base,
        lineno=token.lineno,
        col_offset=token.col_offset + variable_match.start + base_i,
        error=token.error,
    )


def iter_robot_match_as_tokens(
    robot_match: IRobotVariableMatch, relative_index: int = 0, lineno: int = 0
) -> Iterator[IRobotToken]:
    from robot.api import Token

    base = robot_match.base
    assert base is not None
    s = robot_match.string
    if not base:
        base_i = s.find("{") + 1
    else:
        base_i = s.find(base)

    yield Token(
        type="base",
        value=base,
        lineno=lineno,
        col_offset=relative_index + robot_match.start + base_i,
    )

    last_i = base_i + len(base)
    for item in robot_match.items:
        open_char_i = s.find("[", last_i)
        if open_char_i > 0:
            yield Token(
                type="[",
                value="[",
                lineno=lineno,
                col_offset=relative_index + robot_match.start + open_char_i,
            )

            last_i = open_char_i + 1

        if not item:
            item_i = last_i
        else:
            item_i = s.find(item, last_i)

        yield Token(
            type="item",
            value=item,
            lineno=lineno,
            col_offset=relative_index + robot_match.start + item_i,
        )

        last_i = item_i + len(item)

        close_char_i = s.find("]", last_i)
        if close_char_i < 0:
            break

        yield Token(
            type="]",
            value="]",
            lineno=lineno,
            col_offset=relative_index + robot_match.start + close_char_i,
        )

        last_i = close_char_i


def split_token_in_3(
    token: IRobotToken,
    first_token_type: str,
    second_token_type: str,
    third_token_type,
    start_pos,
    end_pos,
) -> Tuple[IRobotToken, IRobotToken, IRobotToken]:
    first = copy_token_replacing(
        token,
        type=first_token_type,
        value=token.value[:start_pos],
    )
    second = copy_token_replacing(
        token,
        type=second_token_type,
        value=token.value[start_pos:end_pos],
        col_offset=token.col_offset + start_pos,
    )

    third = copy_token_replacing(
        token,
        type=third_token_type,
        value=token.value[end_pos:],
        col_offset=second.end_col_offset,
    )

    return first, second, third


def split_token_change_first(
    token: IRobotToken, first_token_type: str, position: int
) -> Tuple[IRobotToken, IRobotToken]:
    prefix = copy_token_replacing(
        token,
        type=first_token_type,
        value=token.value[:position],
    )
    remainder = copy_token_replacing(
        token, value=token.value[position:], col_offset=prefix.end_col_offset
    )
    return prefix, remainder


def split_token_change_second(
    token: IRobotToken, second_token_type: str, position: int
) -> Tuple[IRobotToken, IRobotToken]:
    prefix = copy_token_replacing(
        token,
        value=token.value[:position],
    )
    remainder = copy_token_replacing(
        token,
        value=token.value[position:],
        col_offset=prefix.end_col_offset,
        type=second_token_type,
    )
    return prefix, remainder


def get_library_arguments_serialized(library) -> Optional[str]:
    return "::".join(library.args) if library.args else None


def iter_expression_variables(
    expression_token: IRobotToken,
) -> Iterator[Tuple[IRobotToken, AdditionalVarInfo]]:
    from robot.api import Token

    for tok, var_info in iter_expression_tokens(expression_token):
        if tok.type == Token.VARIABLE:
            yield tok, var_info


class RobotMatchTokensGenerator:
    def __init__(self, token, default_type):
        self.default_type = default_type
        self.token = token
        self.last_gen_end_offset = 0

    def gen_default_type(
        self, until_offset: int
    ) -> Iterable[Tuple[IRobotToken, AdditionalVarInfo]]:
        token = self.token
        if until_offset > self.last_gen_end_offset:
            from robot.api import Token

            val = token.value[self.last_gen_end_offset : until_offset]
            if val.strip():  # Don't generate just for whitespaces.
                yield (
                    Token(
                        self.default_type,
                        val,
                        token.lineno,
                        token.col_offset + self.last_gen_end_offset,
                        token.error,
                    ),
                    AdditionalVarInfo(),
                )
            self.last_gen_end_offset = until_offset

    def gen_tokens_from_robot_match(
        self,
        robot_match: IRobotVariableMatch,
        last_relative_index: int,
        op_type: str = "variableOperator",
        var_type: Optional[str] = None,
    ) -> Iterable[Tuple[IRobotToken, AdditionalVarInfo]]:
        from robot.api import Token
        from robotframework_ls.impl.variable_resolve import is_number_var
        from robotframework_ls.impl.variable_resolve import is_python_eval_var
        from robotframework_ls.impl.variable_resolve import (
            extract_var_name_from_extended_base_name,
        )
        from robotframework_ls.impl.variable_resolve import robot_search_variable

        curr_var_type = var_type
        if curr_var_type is None:
            curr_var_type = Token.VARIABLE

        token = self.token
        if not robot_match.base:
            i = token.value.find("{", robot_match.start + last_relative_index) + 1
        else:
            i = token.value.find(
                robot_match.base, robot_match.start + last_relative_index
            )

        start_offset = robot_match.start + last_relative_index

        yield from self.gen_default_type(start_offset)

        yield (
            Token(
                op_type,
                token.value[robot_match.start + last_relative_index : i],
                token.lineno,
                token.col_offset + start_offset,
                token.error,
            ),
            AdditionalVarInfo(),
        )

        # Base has everything
        base = robot_match.base
        assert base is not None

        first_subvar_match_in_base = robot_search_variable(base)
        has_subvar = bool(
            first_subvar_match_in_base and first_subvar_match_in_base.base
        )

        # Now, we must extract the variable name from the base.
        # ie.: ${a + 1} will provide 'a'.
        var_name_from_base = ""
        if not is_number_var(base) and not is_python_eval_var(base):
            var_name_from_base = extract_var_name_from_extended_base_name(base)
            if var_name_from_base != base:
                while var_name_from_base.endswith((" ", "\t")):
                    var_name_from_base = var_name_from_base[:-1]

        base_or_extended_part = base
        offset = token.col_offset + i
        if var_name_from_base or not base_or_extended_part.strip():
            if not has_subvar or (
                first_subvar_match_in_base
                and first_subvar_match_in_base.start > len(var_name_from_base)
            ):
                base_or_extended_part = base[len(var_name_from_base) :]
                offset += len(var_name_from_base)
                yield (
                    Token(
                        Token.VARIABLE,
                        var_name_from_base,
                        token.lineno,
                        token.col_offset + i,
                        token.error,
                    ),
                    AdditionalVarInfo(
                        robot_match.identifier, extended_part=base_or_extended_part
                    ),
                )

        if base_or_extended_part.strip():
            subvar_tokens = tuple(
                _tokenize_subvars_tokens(
                    Token(
                        curr_var_type,
                        base_or_extended_part,
                        token.lineno,
                        offset,
                        token.error,
                    ),
                    op_type,
                    var_type,
                )
            )

            yield from iter(subvar_tokens)

        j = i + len(base)

        val = token.value[j : robot_match.end + last_relative_index]
        yield (
            Token(
                op_type,
                val,
                token.lineno,
                token.col_offset + j,
                token.error,
            ),
            AdditionalVarInfo(),
        )

        self.last_gen_end_offset = j + len(val)


def _gen_tokens_in_py_expr(
    py_expr,
    expression_token,
) -> Iterator[Tuple[IRobotToken, AdditionalVarInfo]]:
    from tokenize import generate_tokens, NAME, ERRORTOKEN
    from io import StringIO
    from robot.api import Token

    var_type = Token.VARIABLE
    op_type = "variableOperator"

    gen_var_token_info = None
    try:
        for token_info in generate_tokens(StringIO(py_expr).readline):
            if token_info.type == ERRORTOKEN and token_info.string == "$":
                gen_var_token_info = token_info

            elif gen_var_token_info is not None and token_info.type == NAME:
                if gen_var_token_info.start[1] == token_info.start[1] - 1:
                    start_offset = gen_var_token_info.start[1]

                    yield Token(
                        op_type,
                        gen_var_token_info.string,
                        expression_token.lineno,
                        expression_token.col_offset + start_offset,
                        expression_token.error,
                    ), AdditionalVarInfo(context=AdditionalVarInfo.CONTEXT_EXPRESSION)

                    yield Token(
                        var_type,
                        token_info.string,
                        expression_token.lineno,
                        expression_token.col_offset + token_info.start[1],
                        expression_token.error,
                    ), AdditionalVarInfo(
                        "$", context=AdditionalVarInfo.CONTEXT_EXPRESSION
                    )

    except:
        log.exception(f"Unable to evaluate python expression from: {expression_token}")


def iter_expression_tokens(
    expression_token: IRobotToken,
    default_type=None,
) -> Iterator[Tuple[IRobotToken, AdditionalVarInfo]]:
    # See: robot.variables.evaluation.evaluate_expression

    from robotframework_ls.impl.variable_resolve import iter_robot_variable_matches

    if default_type is None:
        default_type = expression_token.ARGUMENT

    expression_to_evaluate: List[str] = []

    robot_matches_and_relative_index = list(
        iter_robot_variable_matches(expression_token.value)
    )

    robot_match = None
    for robot_match, relative_index in robot_matches_and_relative_index:
        expression_to_evaluate.append(robot_match.before)
        expression_to_evaluate.append("1" * (robot_match.end - robot_match.start))

    if robot_match is None:
        after = expression_token.value
    else:
        after = robot_match.after

    if after.strip():
        expression_to_evaluate.append(after)

    python_toks_and_identifiers: list = []
    if expression_to_evaluate:
        expr = "".join(expression_to_evaluate)
        if expr.strip():
            python_toks_and_identifiers.extend(
                _gen_tokens_in_py_expr(expr, expression_token)
            )

    robot_match_generator = RobotMatchTokensGenerator(expression_token, default_type)

    # Now, let's put the vars from python and the robot matches we have in a
    # sorted list so that we can iterate properly.
    from robot.api import Token

    def key(obj):
        # obj is either a tuple(robot match/relative index) or a tuple(Token/var identifier)
        if isinstance(obj[0], Token):
            return obj[0].col_offset

        robot_match: IRobotVariableMatch = obj[0]
        relative_index = obj[1]
        return relative_index + robot_match.start + expression_token.col_offset

    lst = sorted(
        python_toks_and_identifiers + robot_matches_and_relative_index, key=key
    )

    # obj is either a tuple(robot match/relative index) or a tuple(Token/var identifier)
    obj: Any
    for obj in lst:
        if isinstance(obj[0], Token):
            yield from robot_match_generator.gen_default_type(
                obj[0].col_offset - expression_token.col_offset
            )
            yield obj
            robot_match_generator.last_gen_end_offset = (
                obj[0].end_col_offset - expression_token.col_offset
            )

        else:
            yield from robot_match_generator.gen_tokens_from_robot_match(*obj)
    yield from robot_match_generator.gen_default_type(len(expression_token.value))


def is_node_with_expression_argument(node) -> bool:
    if node.__class__.__name__ == "KeywordCall":
        kw_name = node.keyword
        return kw_name and normalize_robot_name(kw_name) == "evaluate"
    else:
        return node.__class__.__name__ in CLASSES_WTH_EXPRESSION_ARGUMENTS


def iter_arguments_from_template(
    stack: Tuple[INode, ...], node: INode
) -> Iterator[NodeInfo]:
    if not stack:
        log.critical(
            "Unable to iterate arguments from template because the stack is empty."
        )
        return

    from robot.api import Token

    if node.type == Token.TEMPLATE:
        for entry in reversed(stack):
            if isinstance_name(entry, "TestCase"):
                for node_info in iter_indexed(entry, "TemplateArguments"):
                    for token in node_info.node.tokens:
                        if token.type == token.ARGUMENT:
                            yield node_info
                            break
        return

    if node.type == Token.TEST_TEMPLATE:
        # We need to collect all the TemplateArguments and skip the ones from
        # tests which have a customized template.
        for node_info in iter_indexed(stack[0], "TestCase"):
            if tuple(iter_indexed(node_info.node, "Template")):
                # i.e.: This test case has a custom template
                continue

            for node_info in iter_indexed(node_info.node, "TemplateArguments"):
                for token in node_info.node.tokens:
                    if token.type == token.ARGUMENT:
                        yield node_info
                        break


def set_localization_info_in_model(ast, localization_info: LocalizationInfo):
    """
    Sets information regarding localization of the AST in the model (File).
    """
    assert (
        ast.__class__.__name__ == "File"
    ), f"Expected File. Found: {ast.__class__.__name__}"

    ast.__localization_info__ = localization_info

    file_weak_ref = weakref.ref(ast)
    for _stack, node in _iter_nodes(ast):
        node.__file_weak_ref__ = file_weak_ref  # type:ignore
        node.__localization_info__ = localization_info  # type:ignore


def get_localization_info_from_model(ast) -> LocalizationInfo:
    """
    Note that the ast should usually be the file or one of the sections right
    below it (such as the settings) where the localization info was set (usually
    it's the stack[0] passed with the node).
    """
    return ast.__localization_info__


def iter_argument_tokens(ast) -> Iterator[INode]:
    for token in ast.tokens:
        if token.type == token.ARGUMENT:
            yield token



from robotframework_ls.impl.protocols import (
    ICompletionContext,
    TokenInfo,
    VarTokenInfo,
    IRobotToken,
    INode,
)
from robocorp_ls_core.lsp import (
    PositionTypedDict,
    SelectionRangeTypedDict,
)
from typing import List, Sequence
from robocorp_ls_core.robotframework_log import get_logger
import itertools

log = get_logger(__name__)


def _empty_range(position: PositionTypedDict) -> SelectionRangeTypedDict:
    line, col = position["line"], position["character"]
    r: SelectionRangeTypedDict = {
        "range": {
            "start": {
                "line": line,
                "character": col,
            },
            "end": {
                "line": line,
                "character": col,
            },
        }
    }
    return r


def _build_variable_range_hierarchy(
    current_token: TokenInfo,
    current_variable: VarTokenInfo,
) -> SelectionRangeTypedDict:
    from robotframework_ls.impl.ast_utils import create_range_from_token

    ret: SelectionRangeTypedDict = {
        "range": create_range_from_token(current_variable.token)
    }

    # The current token is the container of the current variable.
    current_token_selection_range = _build_token_range_hierarchy(current_token)
    if not current_token_selection_range.get("range"):
        return ret

    if current_token_selection_range.get("range") != ret["range"]:
        ret["parent"] = current_token_selection_range
    return ret


def _build_token_range_from_stack(
    ret: SelectionRangeTypedDict,
    current_node: INode,
    stack: Sequence[INode],
    token: IRobotToken,
):
    from robotframework_ls.impl.ast_utils import create_range_from_node

    last: SelectionRangeTypedDict = ret

    for stack_node in itertools.chain((current_node,), reversed(stack)):
        r = create_range_from_node(
            stack_node,
            accept_empty=token.type in (token.EOL, token.EOS, token.SEPARATOR),
        )
        if r is None:
            continue

        if last["range"] == r:  # If it's the same, don't add it.
            continue

        new_range: SelectionRangeTypedDict = {"range": r}
        last["parent"] = new_range
        last = new_range


def _build_token_range_hierarchy(current_token: TokenInfo) -> SelectionRangeTypedDict:
    from robotframework_ls.impl.ast_utils import create_range_from_token

    token = current_token.token
    ret: SelectionRangeTypedDict = {"range": create_range_from_token(token)}
    _build_token_range_from_stack(ret, current_token.node, current_token.stack, token)
    return ret


def selection_range(
    context: ICompletionContext, positions: List[PositionTypedDict]
) -> List[SelectionRangeTypedDict]:
    if not positions:
        return []

    ret: List[SelectionRangeTypedDict] = []

    from robotframework_ls.impl import ast_utils

    ast = context.get_ast()

    for position in positions:
        line = position["line"]
        section = ast_utils.find_section(ast, line)
        if section is None:
            ret.append(_empty_range(position))
            continue

        col = position["character"]

        current_token = ast_utils.find_token(section, line, col)
        if current_token is None:
            ret.append(_empty_range(position))
            continue

        current_variable = ast_utils.find_variable(section, line, col)
        if current_variable is not None:
            ret.append(_build_variable_range_hierarchy(current_token, current_variable))
            continue

        ret.append(_build_token_range_hierarchy(current_token))
    log.info("returning: %s", ret)
    return ret



from typing import List, Dict

from robocorp_ls_core.lsp import (
    LocationTypedDict,
    WorkspaceEditTypedDict,
    TextEditTypedDict,
    MessageType,
)
from robotframework_ls.impl.protocols import ICompletionContext, IKeywordDefinition
from robocorp_ls_core.jsonrpc.exceptions import JsonRpcException
import typing
from robotframework_ls.impl.robot_constants import ROBOT_AND_TXT_FILE_EXTENSIONS
from robocorp_ls_core.robotframework_log import get_logger

log = get_logger(__name__)


def rename(
    completion_context: ICompletionContext, new_name: str
) -> WorkspaceEditTypedDict:
    from robotframework_ls.impl.references import references

    refs: List[LocationTypedDict] = references(
        completion_context, include_declaration=True
    )
    changes: Dict[str, List[TextEditTypedDict]] = {}
    for ref in refs:
        text_range = ref["range"]
        text_edit: TextEditTypedDict = {
            "range": text_range,
            "newText": new_name,
        }

        uri = ref["uri"]

        if uri != completion_context.doc.uri and not uri.lower().endswith(
            ROBOT_AND_TXT_FILE_EXTENSIONS
        ):
            log.info("Skipping rename at: %s", uri)
            continue

        lst = changes.get(uri)
        if lst is None:
            lst = changes[uri] = []

        lst.append(text_edit)

    ret: WorkspaceEditTypedDict = {
        "changes": changes,
    }
    return ret

    # Do we have a usage to require documentChanges instead?
    #
    # changes: Dict[str, List[AnnotatedTextEditTypedDict]] = {}
    # for i, ref in enumerate(refs):
    #     text_range = ref["range"]
    #     text_edit: AnnotatedTextEditTypedDict = {
    #         "range": text_range,
    #         "newText": new_name,
    #         "annotationId": str((i % 2) + 1),
    #     }
    #
    #     uri = ref["uri"]
    #
    #     lst = changes.get(uri)
    #     if lst is None:
    #         lst = changes[uri] = []
    #
    #     lst.append(text_edit)
    #
    # document_changes: List[TextDocumentEditTypedDict] = []
    # for uri, change in changes.items():
    #     text_document_edit: TextDocumentEditTypedDict = {
    #         "textDocument": {"uri": uri, "version": None},
    #         "edits": change,
    #     }
    #     document_changes.append(text_document_edit)
    #
    # ret: WorkspaceEditTypedDict = {
    #     # "changes": changes,
    #     "documentChanges": document_changes,
    #     "changeAnnotations": {"1": "Change 1", "2": "Change 2 not supported?"},
    # }


def prepare_rename(completion_context: ICompletionContext):
    from robotframework_ls.impl.find_definition import find_definition_extended
    from robotframework_ls.impl import ast_utils
    from robotframework_ls.impl.protocols import IVariableDefinition

    lsp_messages = completion_context.lsp_messages

    definition_info = find_definition_extended(completion_context)
    if definition_info:
        for definition in definition_info.definitions:
            if hasattr(definition, "keyword_found"):
                keyword_definition = typing.cast(IKeywordDefinition, definition)
                keyword_found = keyword_definition.keyword_found
                library_name = keyword_found.library_name
                keyword_name = keyword_found.keyword_name

                if library_name:
                    if lsp_messages:
                        lsp_messages.show_message(
                            f"Keyword defined in Library. Only references will be renamed "
                            f"(the '{keyword_name}' definition in '{library_name}' "
                            f"will need to be renamed manually).",
                            MessageType.Warning,
                        )
                else:
                    try:
                        toks = tuple(
                            ast_utils.create_token(keyword_name).tokenize_variables()
                        )
                    except:
                        pass
                    else:
                        for tok in toks:
                            if tok.type == tok.VARIABLE:
                                raise JsonRpcException(
                                    f"Unable to rename '{keyword_name}' (keywords with variables embedded in the name cannot be renamed).",
                                    10,
                                )

                return definition_info.origin_selection_range

            if hasattr(definition, "variable_found"):
                from robotframework_ls.impl.variable_resolve import has_variable

                variable_definition = typing.cast(IVariableDefinition, definition)

                variable_name = variable_definition.variable_found.variable_name
                if has_variable(variable_name):
                    raise JsonRpcException(
                        f"Unable to rename variable definition: '{variable_name}' (variables with variables embedded cannot be renamed).",
                        10,
                    )
                return definition_info.origin_selection_range

    raise JsonRpcException(
        "Unable to rename (could not find keyword nor variable in current position).",
        10,
    )



from robocop.utils.misc import normalize_robot_name


class RunKeywordVariant:
    def __init__(self, name, resolve=1, branches=None, split_on_and=False):
        self.name = normalize_robot_name(name)
        self.resolve = resolve
        self.branches = branches
        self.split_on_and = split_on_and


class RunKeywords(dict):
    def __init__(self, keywords):
        normalized_keywords = {}
        for keyword_variant in keywords:
            normalized_name = normalize_robot_name(keyword_variant.name)
            name_with_lib = f"builtin.{normalized_name}"
            normalized_keywords[normalized_name] = keyword_variant
            normalized_keywords[name_with_lib] = keyword_variant
        super().__init__(normalized_keywords)

    def __setitem__(self, keyword_name, kw_variant):
        normalized_name = normalize_robot_name(keyword_name)
        name_with_lib = f"builtin.{normalized_name}"
        super().__setitem__(normalized_name, kw_variant)
        super().__setitem__(name_with_lib, kw_variant)

    def __getitem__(self, keyword_name):
        normalized_name = normalize_robot_name(keyword_name)
        return super().__getitem__(normalized_name)

    def __missing__(self, keyword_name):
        return None


RUN_KEYWORDS = RunKeywords(
    [
        RunKeywordVariant("Run Keyword"),
        RunKeywordVariant("Run Keyword And Continue On Failure"),
        RunKeywordVariant("Run Keyword And Expect Error", resolve=2),
        RunKeywordVariant("Run Keyword And Ignore Error"),
        RunKeywordVariant("Run Keyword And Return"),
        RunKeywordVariant("Run Keyword And Return If", resolve=2),
        RunKeywordVariant("Run Keyword And Return Status"),
        RunKeywordVariant("Run Keyword And Warn On Failure"),
        RunKeywordVariant("Run Keyword If", resolve=2, branches=["ELSE IF", "ELSE"]),
        RunKeywordVariant("Run Keyword If All Tests Passed"),
        RunKeywordVariant("Run Keyword If Any Tests Failed"),
        RunKeywordVariant("Run Keyword If Test Failed"),
        RunKeywordVariant("Run Keyword If Test Passed"),
        RunKeywordVariant("Run Keyword If Timeout Occurred"),
        RunKeywordVariant("Run Keyword Unless", resolve=2),
        RunKeywordVariant("Run Keywords", split_on_and=True),
        RunKeywordVariant("Repeat Keyword", resolve=2),
        RunKeywordVariant("Wait Until Keyword Succeeds", resolve=3),
    ]
)


def skip_leading_tokens(tokens, break_token):
    for index, token in enumerate(tokens):
        if token.type == break_token:
            return tokens[index:]


def is_token_value_in_tokens(value, tokens):
    return any(value == token.value for token in tokens)


def split_on_token_value(tokens, value, resolve: int):
    """
    Split list of tokens into three lists based on token value.
    Returns tokens before found token, found token + `resolve` number of tokens, remaining tokens.
    """
    for index, token in enumerate(tokens):
        if value == token.value:
            prefix = tokens[:index]
            branch = tokens[index : index + resolve]
            remainder = tokens[index + resolve :]
            return prefix, branch, remainder
    else:
        return [], [], tokens


def iterate_keyword_names(keyword_node, name_token_type):
    tokens = skip_leading_tokens(keyword_node.data_tokens, name_token_type)
    yield from parse_run_keyword(tokens)


def parse_run_keyword(tokens):
    if not tokens:
        return
    yield tokens[0]
    run_keyword = RUN_KEYWORDS[tokens[0].value]
    if not run_keyword:
        return
    tokens = tokens[run_keyword.resolve :]
    if run_keyword.branches:
        if "ELSE IF" in run_keyword.branches:
            while is_token_value_in_tokens("ELSE IF", tokens):
                prefix, branch, tokens = split_on_token_value(tokens, "ELSE IF", 2)
                yield from parse_run_keyword(prefix)
        if "ELSE" in run_keyword.branches and is_token_value_in_tokens("ELSE", tokens):
            prefix, branch, tokens = split_on_token_value(tokens, "ELSE", 1)
            yield from parse_run_keyword(prefix)
            yield from parse_run_keyword(tokens)
            return
    elif run_keyword.split_on_and:
        yield from split_on_and(tokens)
        return
    yield from parse_run_keyword(tokens)


def split_on_and(tokens):
    if not is_token_value_in_tokens("AND", tokens):
        yield from (token for token in tokens)
        return
    while is_token_value_in_tokens("AND", tokens):
        prefix, branch, tokens = split_on_token_value(tokens, "AND", 1)
        yield from parse_run_keyword(prefix)
    yield from parse_run_keyword(tokens)



from robotframework_ls.impl.protocols import (
    ICompletionContext,
    NodeInfo,
    ILibraryDocConversions,
)
from typing import Optional, Any
import typing
from robocorp_ls_core.lsp import HoverTypedDict
from robocorp_ls_core.robotframework_log import get_logger
from robocorp_ls_core.protocols import ActionResultDict

log = get_logger(__name__)


def _collect_html_for_non_library_element(
    ctx: ICompletionContext,
) -> Optional[HoverTypedDict]:
    from robotframework_ls.impl import hover

    return hover.hover(ctx)


def collect_robot_documentation(
    library_name: Optional[str], ctx: ICompletionContext
) -> ActionResultDict:
    try:
        return _collect_robot_documentation(library_name, ctx)
    except Exception as e:
        msg = f"Error collecting robot documentation: {str(e)}"
        log.exception(msg)

        return {
            "success": False,
            "message": msg,
            "result": None,
        }


def _collect_robot_documentation(
    library_name: Optional[str], ctx: ICompletionContext
) -> ActionResultDict:
    from robotframework_ls.impl import ast_utils

    ws = ctx.workspace
    libspec_manager = ws.libspec_manager

    # We need to create a copy (which we'll use for dealing with HTML).
    # Note that we also want to collect the original copy, not the one
    # which was converted to markdown.
    libspec_manager = libspec_manager.create_copy()

    if library_name:
        library_doc_or_error = libspec_manager.get_library_doc_or_error(
            library_name,
            create=True,
            completion_context=ctx,
        )
    else:
        ast = ctx.get_ast()
        from robot.api import Token

        line = ctx.sel.line
        col = ctx.sel.col

        section = ast_utils.find_section(ast, line)
        if not section:
            return {
                "success": False,
                "message": f"No documentation for selection at line: {line}",
                "result": None,
            }
        token_info = ast_utils.find_token(section, line, col)
        if not token_info:
            return {
                "success": False,
                "message": f"No documentation for selection in line: {line}, col: {col}",
                "result": None,
            }

        node_info: NodeInfo[Any] = NodeInfo(token_info.stack, token_info.node)
        if not ast_utils.is_library_node_info(node_info):
            # Ok, no docs for a library, let's get the hover info and provide it.
            ret = _collect_html_for_non_library_element(ctx)
            if not ret:
                return {
                    "success": False,
                    "message": f"No custom documentation available for node: {node_info.node.__class__.__name__} at line: {line}, col: {col} ",
                    "result": None,
                }
            else:
                return {
                    "success": True,
                    "message": None,
                    "result": ret,
                }

        library_name_token = node_info.node.get_token(Token.NAME)
        if library_name_token is None:
            return {
                "success": False,
                "message": f"Unable to get library name for library import in line: {line}, col: {col}.",
                "result": None,
            }

        library_doc_or_error = libspec_manager.get_library_doc_or_error(
            ctx.token_value_resolving_variables(library_name_token),
            create=True,
            completion_context=ctx,
            args=ast_utils.get_library_arguments_serialized(node_info.node),
        )

    library_doc = library_doc_or_error.library_doc
    if not library_doc:
        return {
            "success": False,
            "message": library_doc_or_error.error,
            "result": None,
        }

    try:
        typing.cast(ILibraryDocConversions, library_doc).convert_docs_to_html()
    except Exception as e:
        msg = f"Error converting docs to html: {str(e)}"
        log.exception(msg)

        return {
            "success": False,
            "message": msg,
            "result": None,
        }

    return {
        "success": True,
        "message": None,
        "result": {"libdoc_json": library_doc.to_dictionary()},
    }



from robot.api.parsing import EmptyLine
from robot.parsing.model.blocks import Keyword

from robotidy.disablers import skip_section_if_disabled
from robotidy.transformers import Transformer


class SmartSortKeywords(Transformer):
    """
    Sort keywords in ``*** Keywords ***`` section.

    By default sorting is case insensitive, but keywords with leading underscore go to the bottom. Other underscores are
    treated as spaces.
    Empty lines (or lack of them) between keywords are preserved.

    Following code:

    ```robotframework
    *** Keywords ***
    _my secrete keyword
        Kw2

    My Keyword
        Kw1


    my_another_cool_keyword
    my another keyword
        Kw3
    ```

    Will be transformed to:

    ```robotframework
    *** Keywords ***
    my_another_cool_keyword

    my another keyword
        Kw3


    My Keyword
        Kw1
    _my secrete keyword
        Kw2
    ```

    Default behaviour could be changed using following parameters: ``case_insensitive = True``,
    ``ignore_leading_underscore = False`` and ``ignore_other_underscore = True``.
    """

    ENABLED = False

    def __init__(
        self,
        case_insensitive=True,
        ignore_leading_underscore=False,
        ignore_other_underscore=True,
    ):
        super().__init__()
        self.ci = case_insensitive
        self.ilu = ignore_leading_underscore
        self.iou = ignore_other_underscore

    @skip_section_if_disabled
    def visit_KeywordSection(self, node):  # noqa
        before, after = self.leave_only_keywords(node)
        empty_lines = self.pop_empty_lines(node)
        node.body.sort(key=self.sort_function)
        self.append_empty_lines(node, empty_lines)
        node.body = before + node.body + after
        return node

    @staticmethod
    def pop_empty_lines(node):
        all_empty = []
        for kw in node.body:
            kw_empty = []
            while kw.body and isinstance(kw.body[-1], EmptyLine):
                kw_empty.insert(0, kw.body.pop())
            all_empty.append(kw_empty)
        return all_empty

    @staticmethod
    def leave_only_keywords(node):
        before = []
        after = []
        while node.body and not isinstance(node.body[0], Keyword):
            before.append(node.body.pop(0))
        while node.body and not isinstance(node.body[-1], Keyword):
            after.append(node.body.pop(-1))
        return before, after

    def sort_function(self, kw):
        name = kw.name
        if self.ci:
            name = name.casefold().upper()  # to make sure that letters go before underscore
        if self.ilu:
            name = name.lstrip("_")
        if self.iou:
            index = len(name) - len(name.lstrip("_"))
            name = name[:index] + name[index:].replace("_", " ")
        return name

    @staticmethod
    def append_empty_lines(node, empty_lines):
        for kw, lines in zip(node.body, empty_lines):
            for line in lines:
                kw.body.append(line)



import re
import string
from typing import Optional

from robot.api.parsing import ModelTransformer, Token

from robotidy.disablers import skip_if_disabled, skip_section_if_disabled
from robotidy.exceptions import InvalidParameterValueError
from robotidy.transformers.run_keywords import get_run_keywords
from robotidy.utils import is_token_value_in_tokens, normalize_name, split_on_token_type, split_on_token_value


class RenameKeywords(ModelTransformer):
    """
    Enforce keyword naming.

    Title Case is applied to keyword name and underscores are replaced by spaces.

    You can keep underscores if you set remove_underscores to False:

    ```
    robotidy --transform RenameKeywords -c RenameKeywords:remove_underscores=False .
    ```

    It is also possible to configure `replace_pattern` parameter to find and replace regex pattern. Use `replace_to`
    to set replacement value. This configuration (underscores are used instead of spaces):

    ```
    robotidy --transform RenameKeywords -c RenameKeywords:replace_pattern=^(?i)rename\s?me$:replace_to=New_Shining_Name .
    ```

    will transform following code:

    ```robotframework
    *** Keywords ***
    rename Me
       Keyword Call
    ```

    To:

    ```robotframework
    *** Keywords ***
    New Shining Name
        Keyword Call
    ```

    Use `ignore_library = True` parameter to control if the library name part (Library.Keyword) of keyword call
    should be renamed.
    """

    ENABLED = False

    def __init__(
        self,
        replace_pattern: Optional[str] = None,
        replace_to: Optional[str] = None,
        remove_underscores: bool = True,
        ignore_library: bool = True,
    ):
        self.ignore_library = ignore_library
        self.remove_underscores = remove_underscores
        self.replace_pattern = self.parse_pattern(replace_pattern)
        self.replace_to = "" if replace_to is None else replace_to
        self.run_keywords = get_run_keywords()

    def parse_pattern(self, replace_pattern):
        if replace_pattern is None:
            return None
        try:
            return re.compile(replace_pattern)
        except re.error as err:
            raise InvalidParameterValueError(
                self.__class__.__name__,
                "replace_pattern",
                replace_pattern,
                f"It should be a valid regex expression. Regex error: '{err.msg}'",
            )

    def get_run_keyword(self, kw_name):
        kw_norm = normalize_name(kw_name)
        return self.run_keywords.get(kw_norm, None)

    @skip_section_if_disabled
    def visit_Section(self, node):  # noqa
        return self.generic_visit(node)

    def rename_node(self, token, is_keyword_call):
        values = []
        split_names = token.value.split(".")
        for index, value in enumerate(split_names, start=1):
            if is_keyword_call and self.ignore_library and index != len(split_names):
                values.append(value)
                continue
            if self.replace_pattern is not None:
                value = self.replace_pattern.sub(repl=self.replace_to, string=value)
            if self.remove_underscores and set(value) != {"_"}:
                value = re.sub("_+", " ", value)  # replace one or more _ with one space
            value = value.strip()
            # capitalize first letter of every word, leave rest untouched
            value = "".join([a if a.isupper() else b for a, b in zip(value, string.capwords(value))])
            values.append(value)
        token.value = ".".join(values)

    @skip_if_disabled
    def visit_KeywordName(self, node):  # noqa
        name_token = node.get_token(Token.KEYWORD_NAME)
        if not name_token or not name_token.value:
            return node
        self.rename_node(name_token, is_keyword_call=False)
        return node

    @skip_if_disabled
    def visit_KeywordCall(self, node):  # noqa
        name_token = node.get_token(Token.KEYWORD)
        if not name_token or not name_token.value:
            return node
        # ignore assign, separators and comments
        _, tokens = split_on_token_type(node.data_tokens, Token.KEYWORD)
        self.parse_run_keyword(tokens)
        return node

    def parse_run_keyword(self, tokens):
        if not tokens:
            return
        self.rename_node(tokens[0], is_keyword_call=True)
        run_keyword = self.get_run_keyword(tokens[0].value)
        if not run_keyword:
            return
        tokens = tokens[run_keyword.resolve :]
        if run_keyword.branches:
            if "ELSE IF" in run_keyword.branches:
                while is_token_value_in_tokens("ELSE IF", tokens):
                    prefix, branch, tokens = split_on_token_value(tokens, "ELSE IF", 2)
                    self.parse_run_keyword(prefix)
            if "ELSE" in run_keyword.branches and is_token_value_in_tokens("ELSE", tokens):
                prefix, branch, tokens = split_on_token_value(tokens, "ELSE", 1)
                self.parse_run_keyword(prefix)
                self.parse_run_keyword(tokens)
                return
        elif run_keyword.split_on_and:
            return self.split_on_and(tokens)
        self.parse_run_keyword(tokens)

    def split_on_and(self, tokens):
        if not is_token_value_in_tokens("AND", tokens):
            for token in tokens:
                self.rename_node(token, is_keyword_call=True)
            return
        while is_token_value_in_tokens("AND", tokens):
            prefix, branch, tokens = split_on_token_value(tokens, "AND", 1)
            self.parse_run_keyword(prefix)
        self.parse_run_keyword(tokens)

    @skip_if_disabled
    def visit_SuiteSetup(self, node):  # noqa
        if node.errors:
            return node
        self.parse_run_keyword(node.data_tokens[1:])
        return node

    visit_SuiteTeardown = visit_TestSetup = visit_TestTeardown = visit_SuiteSetup

    @skip_if_disabled
    def visit_Setup(self, node):  # noqa
        if node.errors:
            return node
        self.parse_run_keyword(node.data_tokens[1:])
        return node

    visit_Teardown = visit_Setup



from robot.api.parsing import KeywordCall, Token

try:
    from robot.api.parsing import InlineIfHeader, ReturnStatement
except ImportError:
    InlineIfHeader = None
    ReturnStatement = None

from robotidy.disablers import skip_if_disabled, skip_section_if_disabled
from robotidy.exceptions import InvalidParameterValueError
from robotidy.skip import Skip
from robotidy.transformers import Transformer


class NormalizeSeparators(Transformer):
    """
    Normalize separators and indents.

    All separators (pipes included) are converted to fixed length of 4 spaces (configurable via global argument
    ``--spacecount``).

    You can decide which sections should be transformed by configuring
    ``sections = comments,settings,variables,keywords,testcases`` param.

    To not format documentation configure ``skip_documentation`` to ``True``.
    """

    HANDLES_SKIP = frozenset(
        {"skip_documentation", "skip_keyword_call", "skip_keyword_call_pattern", "skip_comments", "skip_block_comments"}
    )

    def __init__(self, sections: str = None, skip: Skip = None):
        super().__init__(skip=skip)
        self.indent = 0
        self.sections = self.parse_sections(sections)
        self.is_inline = False

    def parse_sections(self, sections):
        default = {"comments", "settings", "testcases", "keywords", "variables"}
        if sections is None:
            return default
        if not sections:
            return {}
        parts = sections.split(",")
        parsed_sections = set()
        for part in parts:
            part = part.replace("_", "")
            if part and part[-1] != "s":
                part += "s"
            if part not in default:
                raise InvalidParameterValueError(
                    self.__class__.__name__,
                    "sections",
                    sections,
                    f"Sections to be transformed should be provided in comma separated "
                    f"list with valid section names:\n{sorted(default)}",
                )
            parsed_sections.add(part)
        return parsed_sections

    def visit_File(self, node):  # noqa
        self.indent = 0
        return self.generic_visit(node)

    def should_visit(self, name, node):
        if name in self.sections:
            return self.generic_visit(node)
        return node

    @skip_section_if_disabled
    def visit_CommentSection(self, node):  # noqa
        return self.should_visit("comments", node)

    @skip_section_if_disabled
    def visit_SettingSection(self, node):  # noqa
        return self.should_visit("settings", node)

    @skip_section_if_disabled
    def visit_VariableSection(self, node):  # noqa
        return self.should_visit("variables", node)

    @skip_section_if_disabled
    def visit_KeywordSection(self, node):  # noqa
        return self.should_visit("keywords", node)

    @skip_section_if_disabled
    def visit_TestCaseSection(self, node):  # noqa
        return self.should_visit("testcases", node)

    def indented_block(self, node):
        self.visit_Statement(node.header)
        self.indent += 1
        node.body = [self.visit(item) for item in node.body]
        self.indent -= 1
        return node

    def visit_TestCase(self, node):  # noqa
        return self.indented_block(node)

    visit_Keyword = visit_While = visit_TestCase  # noqa

    def visit_For(self, node):
        node = self.indented_block(node)
        self.visit_Statement(node.end)
        return node

    def visit_Try(self, node):
        node = self.indented_block(node)
        if node.next:
            self.visit(node.next)
        if node.end:
            self.visit_Statement(node.end)
        return node

    def visit_If(self, node):
        if self.is_inline:  # nested inline if is ignored
            return node
        self.is_inline = InlineIfHeader and isinstance(node.header, InlineIfHeader)
        self.visit_Statement(node.header)
        self.indent += 1
        node.body = [self.visit(item) for item in node.body]
        self.indent -= 1
        if node.orelse:
            self.visit(node.orelse)
        if node.end:
            self.visit_Statement(node.end)
        self.is_inline = False
        return node

    def visit_Documentation(self, doc):  # noqa
        if self.skip.documentation:
            has_pipes = doc.tokens[0].value.startswith("|")
            return self._handle_spaces(doc, has_pipes, only_indent=True)
        return self.visit_Statement(doc)

    def visit_KeywordCall(self, keyword):  # noqa
        if self.skip.keyword_call(keyword):
            return keyword
        return self.visit_Statement(keyword)

    def visit_Comment(self, node):  # noqa
        if self.skip.comment(node):
            return node
        return self.visit_Statement(node)

    def is_keyword_inside_inline_if(self, node):
        return self.is_inline and (
            isinstance(node, KeywordCall) or ReturnStatement and isinstance(node, ReturnStatement)
        )

    @skip_if_disabled
    def visit_Statement(self, statement):  # noqa
        if statement is None:
            return None
        has_pipes = statement.tokens[0].value.startswith("|")
        return self._handle_spaces(statement, has_pipes)

    def _handle_spaces(self, statement, has_pipes, only_indent=False):
        new_tokens = []
        prev_token = None
        for line in statement.lines:
            prev_sep = False
            for index, token in enumerate(line):
                if token.type == Token.SEPARATOR:
                    if prev_sep:
                        continue
                    prev_sep = True
                    if index == 0 and not self.is_keyword_inside_inline_if(statement):
                        token.value = self.formatting_config.indent * self.indent
                    elif not only_indent:
                        if prev_token and prev_token.type == Token.CONTINUATION:
                            token.value = self.formatting_config.continuation_indent
                        else:
                            token.value = self.formatting_config.separator
                else:
                    prev_sep = False
                    prev_token = token
                if has_pipes and index == len(line) - 2:
                    token.value = token.value.rstrip()
                new_tokens.append(token)
        statement.tokens = new_tokens
        self.generic_visit(statement)
        return statement



"""
Main class of Robocop module. Gathers files to be scanned, checkers, parses CLI arguments and scans files.
"""
import os
import sys
from collections import Counter

from robot.api import get_resource_model
from robot.errors import DataError

import robocop.exceptions
from robocop import checkers, reports
from robocop.config import Config
from robocop.files import get_files
from robocop.rules import Message
from robocop.utils import DisablersFinder, FileType, FileTypeChecker, RecommendationFinder, is_suite_templated
from robocop.utils.file_types import check_model_type, get_resource_with_lang


class Robocop:
    """
    Main class for running the checks.

    If you want to run checks with non default configuration create your own ``Config`` and pass it to ``Robocop``.
    Use ``Robocop.run()`` method to start analysis. If ``from_cli`` is set to ``False`` it will return
    list of found issues in JSON format.

    Example::

        import robocop
        from robocop.config import Config

        config = Config()
        config.include = {'1003'}
        config.paths = ['tests\\atest\\rules\\section-out-of-order']

        robocop_runner = robocop.Robocop(config=config)
        issues = robocop_runner.run()

    """

    def __init__(self, from_cli: bool = False, config: Config = None):
        self.files = {}
        self.checkers = []
        self.rules = {}
        self.reports = {}
        self.root = os.getcwd()
        self.config = Config(from_cli=from_cli) if config is None else config
        self.from_cli = from_cli
        if not from_cli:
            self.config.reports.append("json_report")
        self.out = self.set_output()

    def set_output(self):
        """Set output for printing to file if configured. Else use standard output"""
        return self.config.output or None

    def write_line(self, line):
        """Print line using file=self.out parameter (set in `set_output` method)"""
        print(line, file=self.out)

    def reload_config(self):
        """Reload checkers and reports based on current config"""
        self.load_checkers()
        self.config.validate_rule_names(self.rules)
        self.load_reports()
        self.configure_checkers_or_reports()
        self.check_for_disabled_rules()
        self.list_checkers()

    def run(self):
        """Entry point for running scans"""
        self.reload_config()
        self.recognize_file_types()
        self.run_checks()
        self.make_reports()
        if self.config.output and not self.out.closed:
            self.out.close()
        if self.from_cli:
            sys.exit(self.reports["return_status"].return_status)
        else:
            return self.reports["json_report"].issues

    def recognize_file_types(self):
        """
        Pre-parse files to recognize their types. If the filename is `__init__.*`, the type is `INIT`.
        Files with .resource extension are `RESOURCE` type.
        If the file is imported somewhere then file type is `RESOURCE`. Otherwise, file type is `GENERAL`.
        These types are important since they are used to define parsing class for robot API.
        """
        file_type_checker = FileTypeChecker(self.config.exec_dir)
        for file in get_files(self.config):
            if "__init__" in file.name:
                file_type = FileType.INIT
            elif file.suffix.lower() == ".resource":
                file_type = FileType.RESOURCE
            else:
                file_type = FileType.GENERAL
            file_type_checker.source = file
            try:
                resource_parser = file_type.get_parser()
                model = get_resource_with_lang(resource_parser, str(file), self.config.language)
                check_model_type(file_type_checker, model)
                self.files[file] = (file_type, model)
            except DataError:
                print(f"Failed to decode {file}. Default supported encoding by Robot Framework is UTF-8. Skipping file")

        for resource in file_type_checker.resource_files:
            if resource in self.files and self.files[resource][0].value != FileType.RESOURCE:
                self.files[resource] = (
                    FileType.RESOURCE,
                    get_resource_with_lang(get_resource_model, str(resource), self.config.language),
                )

    def run_checks(self):
        for file in self.files:
            if self.config.verbose:
                print(f"Scanning file: {file}")
            model = self.files[file][1]
            found_issues = self.run_check(model, str(file))
            found_issues.sort()
            for issue in found_issues:
                self.report(issue)
        if "file_stats" in self.reports:
            self.reports["file_stats"].files_count = len(self.files)

    def run_check(self, ast_model, filename, source=None):
        disablers = DisablersFinder(filename=filename, source=source)
        if disablers.file_disabled:
            return []
        found_issues = []
        templated = is_suite_templated(ast_model)
        for checker in self.checkers:
            if checker.disabled:
                continue
            found_issues += [
                issue
                for issue in checker.scan_file(ast_model, filename, source, templated)
                if not disablers.is_rule_disabled(issue)
            ]
        return found_issues

    def report(self, rule_msg: Message):
        for report in self.reports.values():
            report.add_message(rule_msg)
        try:
            source_rel = os.path.relpath(os.path.expanduser(rule_msg.source), self.root)
        except ValueError:
            source_rel = rule_msg.source
        self.log_message(
            source=rule_msg.source,
            source_rel=source_rel,
            line=rule_msg.line,
            col=rule_msg.col,
            end_line=rule_msg.end_line,
            end_col=rule_msg.end_col,
            severity=rule_msg.severity.value,
            rule_id=rule_msg.rule_id,
            desc=rule_msg.desc,
            name=rule_msg.name,
        )

    def log_message(self, **kwargs):
        self.write_line(self.config.format.format(**kwargs))

    def load_checkers(self):
        self.checkers = []
        self.rules = {}
        checkers.init(self)

    def list_checkers(self):
        if not (self.config.list or self.config.list_configurables):
            return
        if self.config.list_configurables:
            print(
                "All rules have configurable parameter 'severity'. Allowed values are:"
                "\n    E / error\n    W / warning\n    I / info\n"
            )
        pattern = self.config.list if self.config.list else self.config.list_configurables
        rule_by_id = {rule.rule_id: rule for rule in self.rules.values() if rule.matches_pattern(pattern)}
        rule_by_id = sorted(rule_by_id.values(), key=lambda x: x.rule_id)
        severity_counter = Counter({"E": 0, "W": 0, "I": 0})
        for rule in rule_by_id:
            if self.config.list:
                print(rule)
                severity_counter[rule.severity.value] += 1
            else:
                _, params = rule.available_configurables(include_severity=False)
                if params:
                    print(f"{rule}\n" f"    {params}")
                    severity_counter[rule.severity.value] += 1
        configurable_rules_sum = sum(severity_counter.values())
        plural = "" if configurable_rules_sum == 1 else "s"
        print(
            f"\nAltogether {configurable_rules_sum} rule{plural} with following severity:\n"
            f"    {severity_counter['E']} error rule{'' if severity_counter['E'] == 1 else 's'},\n"
            f"    {severity_counter['W']} warning rule{'' if severity_counter['W'] == 1 else 's'},\n"
            f"    {severity_counter['I']} info rule{'' if severity_counter['I'] == 1 else 's'}.\n"
        )
        print("Visit https://robocop.readthedocs.io/en/stable/rules.html page for detailed documentation.")
        sys.exit()

    def load_reports(self):
        self.reports = reports.get_reports(self.config.reports)
        if self.config.list_reports:
            available_reports = reports.list_reports(self.reports)
            print(available_reports)
            sys.exit()

    def register_checker(self, checker):
        for rule_name, rule in checker.rules.items():
            self.rules[rule_name] = rule
            self.rules[rule.rule_id] = rule
        self.checkers.append(checker)

    def check_for_disabled_rules(self):
        """Check checker configuration to disable rules."""
        for checker in self.checkers:
            if not self.any_rule_enabled(checker):
                checker.disabled = True

    def make_reports(self):
        for report in self.reports.values():
            if report.name == "sarif":
                output = report.get_report(self.config, self.rules)
            else:
                output = report.get_report()
            if output is not None:
                self.write_line(output)

    def any_rule_enabled(self, checker) -> bool:
        for name, rule in checker.rules.items():
            rule.enabled = self.config.is_rule_enabled(rule)
            checker.rules[name] = rule
        return any(msg.enabled for msg in checker.rules.values())

    def configure_checkers_or_reports(self):
        for config in self.config.configure:
            if config.count(":") < 2:
                raise robocop.exceptions.ConfigGeneralError(
                    f"Provided invalid config: '{config}' (general pattern: <rule>:<param>:<value>)"
                )
            rule_or_report, param, value = config.split(":", maxsplit=2)
            if rule_or_report in self.rules:
                rule = self.rules[rule_or_report]
                rule.configure(param, value)
            elif rule_or_report in self.reports:
                self.reports[rule_or_report].configure(param, value)
            else:
                similar = RecommendationFinder().find_similar(rule_or_report, self.rules)
                raise robocop.exceptions.ConfigGeneralError(
                    f"Provided rule or report '{rule_or_report}' does not exist. {similar}"
                )


def run_robocop():
    try:
        linter = Robocop(from_cli=True)
        linter.run()
    except robocop.exceptions.RobotFrameworkParsingError:
        raise
    except robocop.exceptions.RobocopFatalError as err:
        print(f"Error: {err}")
        sys.exit(1)
    except Exception as err:
        message = (
            "\nFatal exception occurred. You can create an issue at "
            "https://github.com/MarketSquare/robotframework-robocop/issues/new/choose . Thanks!"
        )
        err.args = (err.args[0] + message,) + err.args[1:]
        raise err



import traceback


class _Callback(object):

    on_exception = None

    def __init__(self):
        self._callbacks = []

    def register(self, callback):
        self._callbacks.append(callback)

    def __call__(self, *args, **kwargs):
        for c in self._callbacks:
            try:
                c(*args, **kwargs)
            except:
                # If something goes bad calling some callback, just print
                # it and keep on going to the next one.
                traceback.print_exc()
                on_exception = self.on_exception
                if on_exception is not None:
                    on_exception()


class DebugListener(object):
    ROBOT_LISTENER_API_VERSION = 3

    on_start_suite = _Callback()
    on_end_suite = _Callback()
    on_start_test = _Callback()
    on_end_test = _Callback()
    on_log_message = _Callback()
    on_message = _Callback()

    def start_suite(self, data, result):
        self.on_start_suite(data, result)

    def end_suite(self, data, result):
        self.on_end_suite(data, result)

    def start_test(self, data, result):
        self.on_start_test(data, result)

    def end_test(self, data, result):
        self.on_end_test(data, result)

    def log_message(self, message):
        self.on_log_message(message)

    def message(self, message):
        self.on_message(message)

    # def start_keyword(self, data, result):
    #     # This would be nice, but it's not currently supported.
    #
    # def end_keyword(self, data, result):
    #     # This would be nice, but it's not currently supported.


class DebugListenerV2(object):
    ROBOT_LISTENER_API_VERSION = 2

    on_start_keyword = _Callback()
    on_end_keyword = _Callback()
    on_start_suite = _Callback()
    on_end_suite = _Callback()
    on_start_test = _Callback()
    on_end_test = _Callback()
    on_log_message = _Callback()
    on_message = _Callback()

    def start_suite(self, name, attributes):
        self.on_start_suite(name, attributes)

    def end_suite(self, name, attributes):
        self.on_end_suite(name, attributes)

    def start_test(self, name, attributes):
        self.on_start_test(name, attributes)

    def end_test(self, name, attributes):
        self.on_end_test(name, attributes)

    def start_keyword(self, name, attributes):
        self.on_start_keyword(name, attributes)

    def end_keyword(self, name, attributes):
        self.on_end_keyword(name, attributes)

    def log_message(self, message):
        self.on_log_message(message)

    def message(self, message):
        self.on_message(message)


def install_rf_stream_connection(write_message):
    try:
        from robot_out_stream import RFStream
    except ImportError:
        from robotframework_ls import import_robot_out_stream

        import_robot_out_stream()

        from robot_out_stream import RFStream

    def write_str(s):
        write_message({"type": "event", "event": "rfStream", "body": {"msg": s}})

    kwargs = {"--dir": "None", "__write__": write_str}
    rfstream = RFStream(**kwargs)

    DebugListenerV2.on_start_suite.register(rfstream.start_suite)
    DebugListenerV2.on_end_suite.register(rfstream.end_suite)

    DebugListenerV2.on_start_test.register(rfstream.start_test)
    DebugListenerV2.on_end_test.register(rfstream.end_test)

    DebugListenerV2.on_start_keyword.register(rfstream.start_keyword)
    DebugListenerV2.on_end_keyword.register(rfstream.end_keyword)

    DebugListenerV2.on_log_message.register(rfstream.log_message)
    DebugListenerV2.on_message.register(rfstream.message)



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from collections.abc import Mapping


def is_dict_like(item):
    return isinstance(item, Mapping)


def is_string(item):
    return isinstance(item, str)


from collections.abc import MutableMapping
import re


def normalize(string, ignore=(), caseless=True, spaceless=True):
    """Normalizes given string according to given spec.

    By default string is turned to lower case and all whitespace is removed.
    Additional characters can be removed by giving them in ``ignore`` list.
    """
    empty = "" if is_string(string) else b""
    if isinstance(ignore, bytes):
        # Iterating bytes in Python3 yields integers.
        ignore = [bytes([i]) for i in ignore]
    if spaceless:
        string = empty.join(string.split())
    if caseless:
        string = string.lower()
        ignore = [i.lower() for i in ignore]
    # both if statements below enhance performance a little
    if ignore:
        for ign in ignore:
            if ign in string:
                string = string.replace(ign, empty)
    return string


def normalize_whitespace(string):
    return re.sub(r"\s", " ", string, flags=re.UNICODE)


class NormalizedDict(MutableMapping):
    """Custom dictionary implementation automatically normalizing keys."""

    def __init__(self, initial=None, ignore=(), caseless=True, spaceless=True):
        """Initialized with possible initial value and normalizing spec.

        Initial values can be either a dictionary or an iterable of name/value
        pairs. In the latter case items are added in the given order.

        Normalizing spec has exact same semantics as with the :func:`normalize`
        function.
        """
        self._data = {}
        self._keys = {}
        self._normalize = lambda s: normalize(s, ignore, caseless, spaceless)
        if initial:
            self._add_initial(initial)

    def _add_initial(self, initial):
        items = initial.items() if hasattr(initial, "items") else initial
        for key, value in items:
            self[key] = value

    def __getitem__(self, key):
        return self._data[self._normalize(key)]

    def __setitem__(self, key, value):
        norm_key = self._normalize(key)
        self._data[norm_key] = value
        self._keys.setdefault(norm_key, key)

    def __delitem__(self, key):
        norm_key = self._normalize(key)
        del self._data[norm_key]
        del self._keys[norm_key]

    def __iter__(self):
        return (self._keys[norm_key] for norm_key in sorted(self._keys))

    def __len__(self):
        return len(self._data)

    def __str__(self):
        return "{%s}" % ", ".join("%r: %r" % (key, self[key]) for key in self)

    def __eq__(self, other):
        if not is_dict_like(other):
            return False
        if not isinstance(other, NormalizedDict):
            other = NormalizedDict(other)
        return self._data == other._data

    def copy(self):
        copy = NormalizedDict()
        copy._data = self._data.copy()
        copy._keys = self._keys.copy()
        copy._normalize = self._normalize
        return copy

    # Speed-ups. Following methods are faster than default implementations.

    def __contains__(self, key):
        return self._normalize(key) in self._data

    def clear(self):
        self._data.clear()
        self._keys.clear()



import re

from robot.api.parsing import Comment, Token

try:
    from robot.api.parsing import InlineIfHeader
except ImportError:
    InlineIfHeader = None
from robotidy.disablers import skip_if_disabled, skip_section_if_disabled
from robotidy.skip import Skip
from robotidy.transformers import Transformer
from robotidy.utils import ROBOT_VERSION

EOL = Token(Token.EOL)
CONTINUATION = Token(Token.CONTINUATION)


class SplitTooLongLine(Transformer):
    """
    Split too long lines.
    If any line in the keyword call or variable exceeds given length limit (120 by default) it will be
    split:

    ```robotframework
    *** Keywords ***
    Keyword
        Keyword With Longer Name    ${arg1}    ${arg2}    ${arg3}  # let's assume that arg2 is at 120 char
    ```

    To:

    ```robotframework
    *** Keywords ***
    Keyword
        # let's assume that arg2 is at 120 char
        Keyword With Longer Name
        ...    ${arg1}
        ...    ${arg2}
        ...    ${arg3}
    ```

    Allowed line length is configurable using global parameter ``--line-length``:

    ```
    robotidy --line-length 140 src.robot
    ```

    Or using dedicated for this transformer parameter ``line_length``:

    ```
    robotidy --configure SplitTooLongLine:line_length:140 src.robot
    ```

    ``split_on_every_arg`` and ``split_on_every_value`` flags (``True`` by default) controls whether arguments
    and values are split or fills the line until character limit:

    ```robotframework
    *** Test Cases ***
    Test with default split_on_every_arg
        # ${arg1} fits under limit, so it stays in the line
        Keyword With Longer Name    ${arg1}
        ...    ${arg2}    ${arg3}

    Test with split_on_every_arg = False
        # arguments are split
        Keyword With Longer Name
        ...    ${arg1}
        ...    ${arg2}
        ...    ${arg3}
    ```

    Supports global formatting params: ``spacecount`` and ``separator``.
    """

    IGNORED_WHITESPACE = {Token.EOL, Token.CONTINUATION}
    HANDLES_SKIP = frozenset({"skip_keyword_call", "skip_keyword_call_pattern"})

    def __init__(
        self,
        line_length: int = None,
        split_on_every_arg: bool = True,
        split_on_every_value: bool = True,
        skip: Skip = None,
    ):
        super().__init__(skip)
        self._line_length = line_length
        self.split_on_every_arg = split_on_every_arg
        self.split_on_every_value = split_on_every_value
        self.robocop_disabler_pattern = re.compile(
            r"(# )+(noqa|robocop: ?(?P<disabler>disable|enable)=?(?P<rules>[\w\-,]*))"
        )

    @property
    def line_length(self):
        return self.formatting_config.line_length if self._line_length is None else self._line_length

    @skip_section_if_disabled
    def visit_Section(self, node):  # noqa
        return self.generic_visit(node)

    def visit_If(self, node):  # noqa
        if self.is_inline(node):
            return node
        if node.orelse:
            self.generic_visit(node.orelse)
        return self.generic_visit(node)

    @staticmethod
    def is_inline(node):
        return ROBOT_VERSION.major > 4 and isinstance(node.header, InlineIfHeader)

    def should_transform_node(self, node):
        if not self.any_line_too_long(node):
            return False
        # find if any line contains more than one data tokens - so we have something to split
        for line in node.lines:
            count = 0
            for token in line:
                if token.type not in Token.NON_DATA_TOKENS:
                    count += 1
                if count > 1:
                    return True
        return False

    def any_line_too_long(self, node):
        for line in node.lines:
            line = "".join(token.value for token in line)
            line = self.robocop_disabler_pattern.sub("", line)
            line = line.rstrip().expandtabs(4)
            if len(line) >= self.line_length:
                return True
        return False

    def visit_KeywordCall(self, node):  # noqa
        if self.skip.keyword_call(node):
            return node
        if not self.should_transform_node(node):
            return node
        if self.disablers.is_node_disabled(node, full_match=False):
            return node
        return self.split_keyword_call(node)

    @skip_if_disabled
    def visit_Variable(self, node):  # noqa
        if not self.should_transform_node(node):
            return node
        return self.split_variable_def(node)

    @staticmethod
    def join_on_separator(tokens, separator):
        for token in tokens:
            yield token
            yield separator

    @staticmethod
    def split_to_multiple_lines(tokens, indent, separator):
        first = True
        for token in tokens:
            yield indent
            if not first:
                yield CONTINUATION
                yield separator
            yield token
            yield EOL
            first = False

    def split_tokens(self, tokens, line, split_on, indent=None):
        separator = Token(Token.SEPARATOR, self.formatting_config.separator)
        cont_indent = Token(Token.SEPARATOR, self.formatting_config.continuation_indent)
        split_tokens, comments = [], []
        # Comments with separators inside them are split into
        # [COMMENT, SEPARATOR, COMMENT] tokens in the AST, so in order to preserve the
        # original comment, we need a lookback on the separator tokens.
        last_separator = None
        for token in tokens:
            if token.type in self.IGNORED_WHITESPACE:
                continue
            if token.type == Token.SEPARATOR:
                last_separator = token
            elif token.type == Token.COMMENT:
                # AST splits comments with separators, e.g.
                #
                # "# Comment     rest" -> ["# Comment", "     ", "rest"].
                #
                # Notice the third value not starting with a hash - that's what this
                # condition is about:
                if comments and not token.value.startswith("#"):
                    comments[-1].value += last_separator.value + token.value
                else:
                    comments.append(token)
            elif token.type == Token.ARGUMENT:
                if token.value == "":
                    token.value = "${EMPTY}"
                if split_on or not self.col_fit_in_line(line + [separator, token]):
                    line.append(EOL)
                    split_tokens.extend(line)
                    if indent:
                        line = [indent, CONTINUATION, cont_indent, token]
                    else:
                        line = [CONTINUATION, cont_indent, token]
                else:
                    line.extend([separator, token])
        split_tokens.extend(line)
        split_tokens.append(EOL)
        return split_tokens, comments

    def split_variable_def(self, node):
        line = [node.data_tokens[0]]
        tokens, comments = self.split_tokens(node.tokens, line, self.split_on_every_value)
        comments = [Comment([comment, EOL]) for comment in comments]
        node.tokens = tokens
        return (*comments, node)

    def split_keyword_call(self, node):
        separator = Token(Token.SEPARATOR, self.formatting_config.separator)
        cont_indent = Token(Token.SEPARATOR, self.formatting_config.continuation_indent)
        indent = node.tokens[0]

        keyword = node.get_token(Token.KEYWORD)
        # check if assign tokens needs to be split too
        line = [indent, *self.join_on_separator(node.get_tokens(Token.ASSIGN), separator), keyword]
        if not self.col_fit_in_line(line):
            head = [
                *self.split_to_multiple_lines(node.get_tokens(Token.ASSIGN), indent=indent, separator=cont_indent),
                indent,
                CONTINUATION,
                cont_indent,
                keyword,
            ]
            line = []
        else:
            head = []

        tokens, comments = self.split_tokens(
            node.tokens[node.tokens.index(keyword) + 1 :], line, self.split_on_every_arg, indent
        )
        head.extend(tokens)
        comment_tokens = []
        for comment in comments:
            comment_tokens.extend([indent, comment, EOL])

        node.tokens = comment_tokens + head
        return node

    def col_fit_in_line(self, tokens):
        return self.len_token_text(tokens) < self.line_length

    @staticmethod
    def len_token_text(tokens):
        return sum(len(token.value) for token in tokens)



from typing import Iterable

from robot.api.parsing import Token

try:
    from robot.api.parsing import Break, Continue
except ImportError:
    Continue, Break = None, None

from robotidy.disablers import skip_if_disabled, skip_section_if_disabled
from robotidy.transformers import Transformer
from robotidy.utils import after_last_dot, normalize_name, wrap_in_if_and_replace_statement


class ReplaceBreakContinue(Transformer):
    """
    Replace Continue For Loop and Exit For Loop keyword variants with CONTINUE and BREAK statements.

    Following code:

    ```robotframework
    *** Keywords ***
    Keyword
        FOR    ${var}    IN  1  2
            Continue For Loop
            Continue For Loop If    $condition
            Exit For Loop
            Exit For Loop If    $condition
        END
    ```

    will be transformed to:

    ```robotframework
    *** Keywords ***
    Keyword
        FOR    ${var}    IN  1  2
            CONTINUE
            IF    $condition
                CONTINUE
            END
            BREAK
            IF    $condition
                BREAK
            END
        END
    ```
    """

    MIN_VERSION = 5

    def __init__(self):
        super().__init__()
        self.in_loop = False

    def visit_File(self, node):  # noqa
        self.in_loop = False
        return self.generic_visit(node)

    @skip_section_if_disabled
    def visit_Section(self, node):  # noqa
        return self.generic_visit(node)

    @staticmethod
    def create_statement_from_tokens(statement, tokens: Iterable, indent: Token):
        return statement([indent, Token(statement.type), *tokens])

    @skip_if_disabled
    def visit_KeywordCall(self, node):  # noqa
        if not self.in_loop or not node.keyword or node.errors:
            return node
        normalized_name = after_last_dot(normalize_name(node.keyword))
        if "forloop" not in normalized_name:
            return node
        if normalized_name == "continueforloop":
            return self.create_statement_from_tokens(statement=Continue, tokens=node.tokens[2:], indent=node.tokens[0])
        elif normalized_name == "exitforloop":
            return self.create_statement_from_tokens(statement=Break, tokens=node.tokens[2:], indent=node.tokens[0])
        elif normalized_name == "continueforloopif":
            return wrap_in_if_and_replace_statement(node, Continue, self.formatting_config.separator)
        elif normalized_name == "exitforloopif":
            return wrap_in_if_and_replace_statement(node, Break, self.formatting_config.separator)
        return node

    def visit_For(self, node):  # noqa
        self.in_loop = True
        node = self.generic_visit(node)
        self.in_loop = False
        return node

    visit_While = visit_For



from typing import Any, Dict, List, Deque, Optional
from robocorp_ls_core.debug_adapter_core.dap.dap_schema import (
    StartSuiteEvent,
    StartSuiteEventBody,
    StartTestEvent,
    StartTestEventBody,
    EndSuiteEvent,
    EndSuiteEventBody,
    EndTestEvent,
    EndTestEventBody,
)
from collections import namedtuple
from robocorp_ls_core.robotframework_log import get_logger
from robocorp_ls_core import uris
import sys
from robotframework_ls.impl.robot_version import get_robot_major_version

log = get_logger(__name__)


def send_event(event):
    from robotframework_debug_adapter.global_vars import get_global_robot_target_comm

    robot_target_comm = get_global_robot_target_comm()
    if robot_target_comm is not None:
        robot_target_comm.write_message(event)


_SourceInfo = namedtuple("_SourceInfo", "source, lineno, test_name, keyword_name, args")


class _EventsState:
    def __init__(self):
        from collections import deque
        from robotframework_debug_adapter._ignore_failures_in_stack import (
            IgnoreFailuresInStack,
        )

        self._failure_messages: List[str] = []

        self._last_failure_message = ""
        self._last_failure_message_full_stacktrace = ""
        self._last_failure_message_reported_at = None

        self._failed_keywords: List[dict] = []
        self._source_info_stack: Deque[_SourceInfo] = deque()
        self._ignore_failures_in_stack = IgnoreFailuresInStack()
        self._current_suite_filename = None
        self._current_test_filename = None


_global_events_state = _EventsState()


def _get_events_state():
    return _global_events_state


class EventsListenerV3:
    # Note: see https://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html
    # for actual attributes.
    ROBOT_LISTENER_API_VERSION = 3

    def __init__(self):
        # Whenever it's created, reset the global state.
        global _global_events_state
        _global_events_state = _EventsState()

    # start suite/test
    def start_suite(self, data, result) -> None:
        source = data.source
        name = data.name

        tests = []
        for test in data.tests:
            tests.append(test.name)

        state = _get_events_state()
        state._current_suite_filename = source
        state._source_info_stack.append(_SourceInfo(source, None, None, None, None))

        send_event(StartSuiteEvent(StartSuiteEventBody(name, source, tests)))

    def start_test(self, data, result) -> None:
        state = _get_events_state()

        name = data.name
        source = data.source
        state._current_test_filename = source
        lineno = data.lineno
        state._source_info_stack.append(_SourceInfo(source, lineno, name, None, None))

        send_event(StartTestEvent(StartTestEventBody(name, source, lineno)))

    # end suite/test

    def end_suite(self, data, result) -> None:
        state = _get_events_state()
        try:
            state._current_suite_filename = None
            failed_keywords = tuple(reversed(state._failed_keywords))
            state._failed_keywords = []

            send_event(
                EndSuiteEvent(
                    EndSuiteEventBody(
                        data.name,
                        elapsedtime=result.elapsedtime,
                        status=result.status,
                        source=data.source,
                        message=result.message.strip(),
                        failed_keywords=failed_keywords,
                    )
                )
            )
        finally:
            state._source_info_stack.pop()

    def end_test(self, data, result) -> None:
        state = _get_events_state()
        try:
            state._current_test_filename = None
            lst = []

            if state._failure_messages:
                lst.extend(state._failure_messages)

            msg = "\n".join(lst)

            stripped_msg = result.message.strip()
            if stripped_msg not in msg:
                msg = stripped_msg + "\n" + msg

            state._failure_messages = []
            state._last_failure_message = ""

            failed_keywords = tuple(reversed(state._failed_keywords))
            state._failed_keywords = []

            send_event(
                EndTestEvent(
                    EndTestEventBody(
                        data.name,
                        elapsedtime=result.elapsedtime,
                        status=result.status,
                        source=data.source,
                        message=msg.strip(),
                        failed_keywords=failed_keywords,
                    )
                )
            )
        finally:
            state._source_info_stack.pop()


class EventsListenerV2:
    # Note: see https://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html
    # for actual attributes.
    ROBOT_LISTENER_API_VERSION = 2

    # For keywords we're just interested on tracking failures to send when a test/suite finished.
    # Note: we also try to capture the failure through logged messages.

    def __init__(self):
        from robotframework_ls.impl.robot_version import get_robot_major_version

        self._robot_major_version = get_robot_major_version()

    def message(self, message):
        if message["level"] in ("FAIL", "ERROR"):
            return self.log_message(message, skip_error=False)

    def log_message(self, message: Dict[str, Any], skip_error=True) -> None:
        """
        Called when an executed keyword writes a log message.

        message is a dictionary with the following contents:

            message: The content of the message.
            level: Log level used in logging the message.
            timestamp: Message creation time in format YYYY-MM-DD hh:mm:ss.mil.
            html: String yes or no denoting whether the message should be interpreted as HTML or not.

        Not called if the message level is below the current threshold level.
        """

        message_string = message.get("message")
        if not message_string:
            return

        level = message["level"]
        if skip_error and level in ("ERROR",):
            # We do this because in RF all the calls to 'log_message'
            # also generate a call to 'message', so, we want to skip
            # one of those (but note that the other way around isn't true
            # and some errors such as import errors are only reported
            # in 'message' and not 'log_message').
            return

        lst = message_string.splitlines(keepends=False)
        if not lst:
            return

        message_string = "\n".join(lst)
        if not message_string.endswith("\n"):
            message_string += "\n"

        state = _get_events_state()

        if state._ignore_failures_in_stack.ignore():
            return

        from robotframework_debug_adapter.message_utils import (
            extract_source_and_line_from_message,
        )

        level = message.get("level")

        source = None
        lineno = None
        test_name = None

        source_and_line = extract_source_and_line_from_message(message_string)
        if source_and_line is not None:
            source, lineno = source_and_line

        if state._source_info_stack:
            from robotframework_debug_adapter import file_utils

            source_info: _SourceInfo

            check = []
            check.append(
                state._source_info_stack[-1]
            )  # Try the real pos, if not possible...
            try:
                check.append(state._source_info_stack[1])  # Use the test
            except IndexError:
                pass
            try:
                check.append(state._source_info_stack[0])  # Use the suite
            except IndexError:
                pass

            for source_info in check:
                test_name = source_info.test_name  # May be None.

                if source is None:
                    source = source_info.source
                    if source is None:
                        continue
                    source = file_utils.get_abs_path_real_path_and_base_from_file(
                        source
                    )[0]

                    try:
                        lineno = source_info.lineno
                    except AttributeError:
                        pass

                if source:
                    break

        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import (
            LogMessageEvent,
            LogMessageEventBody,
        )

        is_failure = message["level"] in ("FAIL", "ERROR")  # FAIL/WARN/INFO/DEBUG/TRACE

        base_message_string = message_string
        if is_failure:
            message_string = f"{message_string}\n{self._print_stack_trace(state)}"

        send_event(
            LogMessageEvent(
                body=LogMessageEventBody(
                    source=source,
                    lineno=lineno,
                    message=f"{message_string}",
                    level=level,
                    testOrSuiteSource=state._current_test_filename
                    or state._current_suite_filename,
                    testName=test_name,
                )
            )
        )

        if is_failure:
            state._failure_messages.append(message_string)
            state._last_failure_message = base_message_string
            state._last_failure_message_full_stacktrace = message_string
            state._last_failure_message_reported_at = (source, lineno)

    def _print_stack_trace(self, state):
        from robotframework_debug_adapter import file_utils
        import linecache

        stack_trace = []
        stack_trace.append("Traceback (most recent call last):")
        found = len(state._source_info_stack) > 0

        source_info: _SourceInfo
        for source_info in state._source_info_stack:
            source = source_info.source
            keyword_name = source_info.keyword_name
            found = True
            if not source:
                stack_trace.append(f"  <unknown source> ({keyword_name})")
                continue
            source = file_utils.get_abs_path_real_path_and_base_from_file(source)[0]
            lineno = source_info.lineno

            line = ""
            try:
                line = linecache.getline(source, lineno).strip()
            except:
                pass

            # We need to put the file in uri format#lineno.
            # see: https://github.com/microsoft/vscode/issues/150702
            source = uris.from_fs_path(source)
            if keyword_name:
                stack_trace.append(f"  {source}#{lineno} ({keyword_name})")
                if line:
                    stack_trace.append(f"    {line}")

            else:
                test_name = source_info.test_name
                if test_name:
                    found = True
                    stack_trace.append(f"  {source}#{lineno} [{test_name}]")
                    if line:
                        stack_trace.append(f"    {line}")

        if not found:
            return ""
        stack_trace.append("")

        return "\n".join(stack_trace)

    def start_keyword(self, name: str, attributes: Dict[str, Any]) -> None:
        state = _get_events_state()
        state._ignore_failures_in_stack.push(attributes.get("kwname", ""))
        source: Optional[str] = attributes.get("source")
        lineno: Optional[int] = attributes.get("lineno")
        if not source:
            # HACK for RF 3: try to get the location (since it's not available).
            if get_robot_major_version() < 4:
                f: Optional[Any]
                f = sys._getframe()
                while f is not None:
                    if f.f_code.co_name == "run_step":
                        step = f.f_locals.get("step")
                        if step is not None:
                            try:
                                source = step.source
                                lineno = step.lineno
                            except AttributeError:
                                pass
                        break  # Break when run_step is found anyways.

                    f = f.f_back

        test_name = None
        if state._source_info_stack:
            # Keep same test name
            test_name = state._source_info_stack[-1].test_name
        state._source_info_stack.append(
            _SourceInfo(source, lineno, test_name, name, attributes.get("args", ()))
        )

    def end_keyword(self, name: str, attributes: Dict[str, Any]) -> None:
        state = _get_events_state()
        try:
            source_info: _SourceInfo = state._source_info_stack.pop()
        except:
            log.exception("Error in state._source_info_stack.pop()")
            return
        try:
            if state._ignore_failures_in_stack.ignore():
                return

            status = attributes.get("status")

            # Status could be PASS, FAIL, SKIP or NOT RUN
            if status == "FAIL":
                key = (source_info.source, source_info.lineno)
                if key == state._last_failure_message_reported_at:
                    msg = "[FAIL] " + state._last_failure_message_full_stacktrace
                else:
                    msg = "[FAIL INSIDE] " + state._last_failure_message
                state._failed_keywords.append(
                    {
                        "name": source_info.keyword_name,
                        "source": source_info.source,
                        "lineno": source_info.lineno,
                        "message": msg,
                    }
                )

        finally:
            state._ignore_failures_in_stack.pop()
